"""This module implements the PID (proportional integral derivative) controller."""

# Mathematics
from numpy import dot
from numpy import ndarray
from numpy import zeros_like

# Data modelling
from pandas import concat
from pandas import DataFrame


class Pid:

    """The PID controller.

    The proportional-integral-derivative controller (PID) is an industriy-standard feedback control loop.
    This controller responds proportionally to the error, i.e. deviation of the desired state,
    its time derivative and integral.

    Examples:
        First, import some helper functions from numpy.

        >>> from numpy import array

        We then setup the Pid controller with some control constants.

        >>> controller = Pid(
        ...     array([0.5]),
        ...     array([0.0]),
        ...     array([0.0]),
        ...     0.1,
        ... )

        We then specify an initial and desired state as well as the current state derivative.

        >>> initial = array([5.0])
        >>> desired = array([0.0])
        >>> derivative = array([0.0])

        Finally, we retrieve a control signal from the Pid based on the values we just set.

        >>> signal = controller.control(desired, initial, derivative)

    Args:
        P: Proportional control constant ``(n,)``
        I: Integral control constant ``(n,)``
        D: Derivative control constant ``(n,)``
        dt: Time between measurements
        keep_trace: Whether to store a trace of control signals, states, etc.

    References:
        - https://en.wikipedia.org/wiki/PID_controller
    """

    # In this context, we reproduce a common PID notation
    # pylint: disable=invalid-name, too-many-arguments

    def __init__(self, P: ndarray, I: ndarray, D: ndarray, dt: float, keep_trace: bool = False):  # noqa: E741
        # Controller specification
        self.P = P
        self.I = I  # noqa: E741
        self.D = D
        self.dt = dt

        # Error summation field
        self.summed_error = zeros_like(P).transpose()

        # Objects for process tracing
        self.keep_trace = keep_trace
        self.process = DataFrame(
            columns=[
                "desired",
                "state",
                "state_derivative",
                "error",
                "summed_error",
                "proportional",
                "integral",
                "derivative",
                "control_signal",
            ]
        )

    def control(self, desired: ndarray, state: ndarray, state_derivative: ndarray) -> ndarray:
        """Compute the control signal based on proportional, integral and derivative terms.

        Args:
            desired: The desired state
            state: The current state
            state_derivative: The current state's derivative

        Returns:
            The control signal
        """

        # Compute errors
        error = desired - state
        self.summed_error += self.dt * error

        # Compute PID values
        proportional = dot(self.P, error)
        integral = dot(self.I, self.summed_error)
        derivative = dot(self.D, state_derivative)

        # Compute control signal
        control_signal: ndarray = proportional + integral - derivative

        # Append control step to process trace
        if self.keep_trace:
            new = DataFrame(
                {
                    "desired": (desired.copy(),),
                    "state": (state.copy(),),
                    "state_derivative": (state_derivative.copy(),),
                    "error": (error.copy(),),
                    "summed_error": (self.summed_error.copy(),),
                    "proportional": (proportional.copy(),),
                    "integral": (integral.copy(),),
                    "derivative": (derivative.copy(),),
                    "control_signal": (control_signal.copy(),),
                },
            )
            self.process = concat([self.process, new], ignore_index=True)

        # Return result
        return control_signal

    def reset(self) -> None:
        """Resets the filter's memory, i.e. set the error integral to zero and empty the process trace."""

        self.summed_error = zeros_like(self.P).transpose()
        self.process = self.process[0:0]
