"""This module adds support for a Spatialite database (SQLite DB with extra modules).

This module requires the *libsqlite3-mod-spatialite* system dependency.
The database allows for storage and efficient retrieval via spatial indexing of elements.

References:
    - `General information <https://www.gaia-gis.it/gaia-sins/spatialite-manual-2.3.1.html>`__
    - `The website of Spatialite <https://www.gaia-gis.it/fossil/libspatialite/index/>`__
    - `The website of Spatialite and friends <https://www.gaia-gis.it/gaia-sins/>`__
    - `Cookbook, Chapter "Spatial Indexing support"
      <https://www.gaia-gis.it/gaia-sins/spatialite-cookbook-5/cookbook_topics.06.html#topic_Spatial_Indexing_support>`__
    - `Cookbook, Chapter "Creating a well designed DB"
      <https://www.gaia-gis.it/gaia-sins/spatialite-cookbook-5/cookbook_topics.03.html#topic_Creating_a_well_designed_DB>`__
    - `SQL functions reference list <https://www.gaia-gis.it/gaia-sins/spatialite-sql-5.0.0.html>`__
"""

# Python standard
from contextlib import closing
from contextlib import contextmanager
from math import degrees
import random
import string
from warnings import warn

# Database interface
import sqlite3
from sqlite3 import Connection

# Typing
from typing import cast
from typing import Generator
from typing import Iterable
from typing import Iterator
from typing import Optional

# Shapely for internal abstraction
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.geometry import Polygon
import shapely.wkb

# Planning primitives
from pyrate.plan.geometry import Direction
from pyrate.plan.geometry import LocationType
from pyrate.plan.geometry import PolarGeometry
from pyrate.plan.geometry import PolarLocation
from pyrate.plan.geometry import PolarPolygon
from pyrate.plan.geometry import PolarRoute

# Geospatial helpers
from pyrate.plan.geometry.helpers import difference_latitude
from pyrate.plan.geometry.helpers import difference_longitude
from pyrate.plan.geometry.helpers import meters2rad

# Import this to enable GDAL/libgeos exceptions if it has not already happened
from . import s57_files as tmp_import

del tmp_import


class SpatialiteDatabase:

    """Allows for IO with the *Spatialite* SQLite database containing obstacles.

    Reading of entries from the database is implemented using generators, i.e. the elements are retrieved
    one by one as they are consumed by the caller. While this allows for the processing of large amounts of
    data in constant memory, it also keeps the cursor to the database open until all elements have been
    consumed. To consume all of the rows at once simply wrap it into the list constructor like this:
    ``all_as_list = list(database.read_all_obstacles())``. Note that only the parsing to Pyrate primitives is
    done lazily, while the actual database reading happens eagerly.

    Internally, a spatial index is used for fast *retrieval* of obstacles given geometric constrains.
    For example, this makes queries for all obstacles in a given bounding boxes take time roughly
    proportional to the result set, and not the total size of the database.
    Some real-world benchmarks can be obtained from the script :ref:`script-benchmark_db_and_projections`
    and are discussed in :ref:`design-decisions-local-projections`.
    See ``SpatialiteDatabase._CREATE_TABLES_SQL_STATEMENT`` for details on the structure of the
    database. The longitude is always the first/the X component of the two-dimensional geometries.

    *QGIS* can natively open the created databases for visual inspection. It's very efficient too.

    A single polygon in the dabase might get split into multiple ones in a query due to clipping. A unique
    index is maintained with best-effort by adding subsequent numbers to the other slices of the same polygon.
    This assumes that indices are somewhat uniformly distributed and not sequential numbers.

    Examples:
        First, let us create some polygons to be stored (PolarPoint and PolarRoute would also work):

        >>> from pyrate.plan.geometry import PolarLocation, PolarPolygon, LocationType
        >>> locations = [PolarLocation(50, 50), PolarLocation(50, 51), PolarLocation(51, 51)]
        >>> polygon_1 = PolarPolygon(locations=locations, name="A Polygon, YaY!")
        >>> polygon_2 = PolarPolygon(locations=locations, name="Some Name", identifier=42,
        ...                          location_type=LocationType.LAND)
        >>> polygons = [polygon_1, polygon_2]
        >>> polygons  #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        [PolarPolygon(locations=[...], name="A Polygon, YaY!"),
         PolarPolygon(locations=[...], location_type=LocationType.LAND, name="Some Name", identifier=42)]

        Then, you can simply store and then retrieve some polygons.
        Note, that you have to call :meth:`SpatialiteDatabase.close` after using it or use it as a context
        manager, as shown here.

        >>> from pyrate.common.charts import SpatialiteDatabase
        >>> with SpatialiteDatabase(":memory:") as database:
        ...     print(len(database))
        ...     database.write_geometries(polygons)
        ...     # We need to wrap it into a call to `list()` to evaluate the generator returned by
        ...     # `read_obstacles_around` while the database is still open
        ...     read = list(database.read_geometries_around(locations[0], radius=200_000))  # 200 km
        ...     assert len(database) == len(read)
        ...     print(len(database))
        0
        2
        >>> # The database does not guarantee an order of the result set
        >>> sort_by_name = lambda geometry: geometry.name
        >>> read = list(sorted(read, key=sort_by_name))
        >>> read  #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        [PolarPolygon(locations=[...], name="A Polygon, YaY!", identifier=1),
         PolarPolygon(locations=[...], location_type=LocationType.LAND, name="Some Name", identifier=42)]
        >>> read == polygons
        False
        >>> # This is due to the first polygon now being given a new unused identifier
        >>> read[0].identifier
        1
        >>> # So we reset it here for the comparison to succeed
        >>> read[0].identifier = None
        >>> # Now, they should be almost equal (except for floating-point approximations)
        >>> polygon_1.equals_almost_congruent(read[0])
        True
        >>> polygon_2.equals_almost_congruent(read[1])
        True

        A full example application can be found in the script :ref:`script-benchmark_db_and_projections`.

    Possible extensions:
        - Allow for retrieving from arbitrary PolarPolygon/bounding box. See ``_read_obstacles_clipped``.
        - The method :meth:`~read_obstacles_around` could be easily extended/complemented to support ellipses.
          However, rotation of that eclipse would only make this useful, and adding that appears to be tricky.

    Note:
        Use with ``storage_path=":memory:"`` (see example above) to open a ephemeral database that resides in
        RAM. This works when only a single databse user (like a :class:`~SpatialiteDatabase` instance) will
        access it.
        Otherwise, passing ``"file::memory:?cache=shared"`` as a file will allow the same database to be
        accessed by multiple different users within the same process. Both are useful for (unit-)testing too.
        When passing extra options like ``"file::memory:?cache=shared"``,
        you will also have to pass ``uri=True`` to :class:`~SpatialiteDatabase`
        such that the parameters do not get mistaken for being part of the an actual file name:
        >>> with SpatialiteDatabase("file::memory:?cache=shared", uri=True) as database:
        ...     print(len(database))
        0

    Args:
        storage_path: the path where to look for the database file, usually ending with ``.sqlite``
        issue_create_statement: tries to create the table(s) and indices if not yet existing:
            this can be safely let enabled as existing tables make this a NO-OP
        kwargs: additional parameters to be passed to the database creation, see :class:`sqlite3.Connection`

    Raises:
        IOError: When the data base file cannot be accessed
        RuntimeError: If the Spatialite extension (*libsqlite3-mod-spatialite*) cannot be loaded
    """

    #: The Spatial Reference System Identifier used for storing objects
    #: this is the widely used WGS 84; see: https://spatialreference.org/ref/epsg/4326/
    _SRID: int = 4326

    def __init__(self, storage_path: str, issue_create_statement: bool = True, **kwargs) -> None:
        # This may raise an IOError:
        self._connection: Connection = sqlite3.connect(storage_path, **kwargs)

        try:
            # load the spatialite module
            self._connection.load_extension("mod_spatialite.so")
        except sqlite3.OperationalError as error:  # pragma: no cover
            raise RuntimeError(
                "Cannot load the spatialite extension. Is it installed (see installation instructions)? "
                f"Error was: {error}"
            ) from error

        if issue_create_statement:
            self.create_tables()

    def create_tables(self, table_name: str = "Obstacles") -> None:
        """This creates the table(s) and indices in the database (if they do not exist yet).

        See the module documentation of :mod:`pyrate.common.charts.db` for more information.

        Args:
            table_name: The name of the table to initialize
        """
        # Check if the table "Obstacles" is present
        check = f"SELECT COUNT(1) FROM SQLITE_MASTER WHERE name = '{table_name}'"
        with closing(self._connection.execute(check)) as cursor:
            count = cast(int, cursor.fetchone()[0])  # This needs to be cast as the result is untyped

        if count == 0:
            # It is not present, so we initialize the database here

            statement = f"""
            CREATE TABLE IF NOT EXISTS '{table_name}' (
                id INTEGER PRIMARY KEY NOT NULL,
                location_type TINYINT unsigned NOT NULL DEFAULT 0,
                name VARCHAR CHARACTER DEFAULT NULL,
                CHECK (location_type <= {LocationType.max_value()})
            );
            CREATE INDEX IF NOT EXISTS by_location_type ON '{table_name}' (location_type);

            SELECT InitSpatialMetaDataFull();
            SELECT AddGeometryColumn(
              '{table_name}',
              'geometry',
              {SpatialiteDatabase._SRID},
              'GEOMETRY', -- just as fast as e.g. POLYGON but more flexible
              'XY',
              TRUE -- set to NOT NULL
            );
            SELECT CreateSpatialIndex(
              '{table_name}',
              'geometry'
            );

            -- can only be done after column is added; but is not supported by SQLite
            -- ALTER TABLE '{table_name}' ADD CHECK (IsValid(geometry));
            """

            with self._connection:  # auto-commits at the end
                with self.disable_synchronization():
                    self._connection.executescript(statement).close()

    @contextmanager
    def disable_synchronization(self) -> Iterator[None]:
        """Temporarily disables file system synchronization for consistency to increase write performance.

        To quote the `documentation of SQLite <https://www.sqlite.org/pragma.html#pragma_synchronous>`__:

            "With synchronous OFF (0), SQLite continues without syncing as soon as it has handed data off to
            the operating system. If the application running SQLite crashes, the data will be safe, but the
            database might become corrupted if the operating system crashes or the computer loses power before
            that data has been written to the disk surface. On the other hand, commits can be orders of
            magnitude faster with synchronous OFF."

        """
        self._connection.execute("PRAGMA synchronous=OFF").close()
        yield
        self._connection.execute("PRAGMA synchronous=ON").close()

    def copy_contents_to_database(self, file_path: str, update: bool = False, **kwargs) -> None:
        """Dumps the content of this obstacle database to a new one.

        This can be useful in cases where one wants to first create an extremely fast in-memory database and
        later copy it to a file on disk.

        Args:
            file_path: the path of the other database
            update: is set to ``True``, update/replace on conflicting identifier; else raise an error in that
                    case
            kwargs: additional parameters to be passed to the database creation, see attribute ``kwargs`` of
                    :class:`~SpatialiteDatabase`

        Raises:
            sqlite3.IntegrityError: If a duplicate key should have been inserted and ``update`` was set to
                                    ``False``
        """
        # init the other database
        with SpatialiteDatabase(file_path, **kwargs):
            pass

        database_name = _random_name()

        command = "REPLACE" if update else "INSERT"
        statements = f"""
        ATTACH '{file_path}' AS {database_name};
        {command} INTO {database_name}.Obstacles SELECT * FROM main.Obstacles;
        DETACH {database_name};
        """

        with self._connection:  # auto-commits at the end
            self._connection.executescript(statements).close()

    def simplify_contents(self, simplify_tolerance: float) -> None:
        """Simplifies all geometries within the database. Always runs :meth:`~vacuum` afterwards.

        Args:
            simplify_tolerance: the tolerance within all new points shall lie wrt. to the old ones, in meters,
                                non-negative. Set to zero to disable.

        Further ideas:
            - Keep topology between objects, not just within them, e.g. see
              `this blog post <https://trac.osgeo.org/postgis/wiki/UsersWikiSimplifyPreserveTopology>`__.
        """
        assert simplify_tolerance >= 0, "tolerance must be non-negative"

        if simplify_tolerance > 0:
            tolerance_degrees = degrees(meters2rad(simplify_tolerance))
            statement = (
                f"UPDATE Obstacles SET geometry = ST_SimplifyPreserveTopology(geometry, {tolerance_degrees})"
            )

            with self._connection:  # auto-commits at the end
                self._connection.execute(statement).close()

        self.vacuum()

    def vacuum(self) -> None:
        """Defragments the database. This is useful after deleting or shrinking many entries."""
        with self._connection:  # auto-commits at the end
            self._connection.execute("VACUUM").close()

    def write_geometry(
        self, geometry: PolarGeometry, update: bool = False, raise_on_failure: bool = True
    ) -> None:
        """Alias for ``write_obstacles([obstacle], update)``. See :meth:`~write_obstacles`.

        Args:
            geometry: the geometry to place into the database, identified by
                      its :attr:`~pyrate.plan.geometry.Geospatial.identifier`
            update: see :meth:`~write_obstacles`
            raise_on_failure: see :meth:`~write_obstacles`
        """

        return self.write_geometries([geometry], update=update, raise_on_failure=raise_on_failure)

    def write_geometries(
        self, geometries: Iterable[PolarGeometry], update: bool = False, raise_on_failure: bool = True
    ) -> None:
        """Writes geometries into the database.

        All geometries are only identified by their identifier as a primary key.
        If ``update is True``, any existing geometries with the same IDs will be updated/replaced.
        If ``update is False``, an exception is raised if duplicate keys are to be inserted.

        Args:
            geometries:
                The geometries to place into the database, identified by their
                :attr:`~pyrate.plan.geometry.Geospatial.identifier`.
            update:
                If set to ``True``, update/replace on conflicting identifiers;
                else raise an error in that case.
                If set to ``True``, no guarantees about inserts can be made (see :class:`ValueError` below).
            raise_on_failure:
                If set to ``False`` suppress the :class:`ValueError` below and instead print a warning.

        Raises:
            sqlite3.IntegrityError: If a duplicate key should have been inserted and ``update`` was set to
                                    ``False``
            sqlite3.IntegrityError: If a value was not within the constraints; should never happen if all
                                    :class:`~pyrate.plan.geometry.polygon.PolarPolygon` were created properly
            ValueError:
                If the provided geometries are not valid according to *spatialite* and could not be repaired.
                However, any valid geometries will have been inserted by then.
                Also, this is only possible to be checked if ``update`` is set to ``False``.
                Else, incomplete inserts will simply be ignored.
                Only very basic cleanup is attempted.
                Suppressed if ``raise_on_failure is False``.
        """

        count_before = self.count_geometries()

        # Build query
        command = "REPLACE" if update else "INSERT"
        statement = f"""
            WITH _insert_temp(id,location_type,name,geometry)
            AS (VALUES (?,?,?,SanitizeGeometry(GeomFromWKB(?,{SpatialiteDatabase._SRID}))))

            {command} INTO Obstacles
            SELECT * FROM _insert_temp WHERE IsValid(geometry)
            """

        # Convert data
        data = [(g.identifier, g.location_type, g.name, to_wkb(g)) for g in geometries]

        # Execute statement
        with self._connection:  # auto-commits at the end
            self._connection.executemany(statement, data).close()

        # TODO(Felix.Divo):
        # We want to notify the user if the insert was incomplete, i.e. if a geometry was invalid
        # (1) `cursor.rowcount` from executemany() does not work since it returns -1
        # (2) `cursor.lastrowid` contains only the last ID, so we cannot use that
        # (3) Appending `RETURNING id` causes exceptions when used with executemany():
        #      "sqlite3.ProgrammingError: executemany() can only execute DML statements"
        # (4) So we do the stupid thing: Count before and after. Does not work with ``update=True``!
        #     This could also cause problems with concurrency.
        # (5) One could also repair by Buffer(), though that might not be what is desired,
        #     cf. https://shapely.readthedocs.io/en/stable/manual.html#object.buffer.

        # Make sure that all rows were valid
        missing = count_before + len(data) - self.count_geometries()
        if not update and missing > 0:
            message = f"{missing} of the {len(data)} geometries were invalid"
            if raise_on_failure:
                raise ValueError(message)
            warn(message)

    def read_all(
        self, only_location_type: Optional[LocationType] = None
    ) -> Generator[PolarGeometry, None, None]:
        """Read all stored geometries, optionally filtered by a type.

        Args:
            only_location_type: get only geometries of that type, if not set to ``None``

        Yields:
            The geometries as read from the database.
        """
        yield from self._read_geometries(
            geometry_column="geometry",  # no transformation, just select the column
            only_location_type=only_location_type,
        )

    def read_geometries_around(
        self,
        around: PolarLocation,
        radius: float = 10_000.0,
        only_location_type: Optional[LocationType] = None,
    ) -> Generator[PolarGeometry, None, None]:
        """Reads and clips geometries in a given radius around some location.

        The geometries are clipped to the circle given by the location ``around`` and ``radius``. This means
        that any parts stretching beyond the circles are not returned and the geometry approximately follows
        the outer circle in such cases. If the ellipse if deformed at (very) high/low latitudes, the clipping
        area is selected such that at least all geometries in the defined area are included, and possibly some
        more.

        Note:
            This method internally uses an ellipse as the clipping area even for circular clipping areas in
            order to account for some distortions of polar coordinates at high latitudes. Also, keep in mind
            that clipping/selecting for inclusion in the result set is not a perfect operation, as the
            clipping area is internally discretized to a geometry. It thus has corners, and is not smooth
            like an ellipse in the mathematical sense.

        Args:
            around: The point around which obstacles are to be extracted; assumed to be in degrees
            radius: The radius around *around* to retrieve items from in meters; default: 10 000 m.
                    The behaviour is unspecified if it is very large, like more than 1 000 000 m.
                    It must be at least zero.
            only_location_type: Get only obstacles of that type, if not set to ``None``

        Yields:
            The geometries as read from the database.
        """

        # Safety assertions on given radius
        assert radius >= 0.0, "the radius must be non-negative"
        assert (
            radius <= 1_000_000
        ), "radius > 1_000_000; see docs; this is not a fundamental restriction but it is untested"

        # The distance in polar coordinate is always equal when going either west or east
        east_west, _ = around.translate(Direction.East, radius)
        longitudal_radius = difference_longitude(around.longitude, east_west.longitude)

        # The above might be different when going north or south
        # To disambiguate, we take the larger of the two distances as the radius
        north, _ = around.translate(Direction.North, radius)
        south, _ = around.translate(Direction.South, radius)

        latitudal_radius = max(
            difference_latitude(around.latitude, north.latitude),
            difference_latitude(around.latitude, south.latitude),
        )

        # Place a corner of the discretized ellipse every 24 degrees,
        # i.e. turn it into a polygon with 15 corners
        every_degrees = 24

        # MakeEllipse(...) takes the two radii in lat/long direction in degrees and returns a line string
        clipping_area = (
            f"MakePolygon(MakeEllipse({around.longitude}, {around.latitude}, "
            f"{longitudal_radius}, {latitudal_radius}, {SpatialiteDatabase._SRID}, {every_degrees}))"
        )

        yield from self._read_geometries_clipped(clipping_area, only_location_type)

    def _read_geometries_clipped(
        self, clipping_area: str, only_location_type: Optional[LocationType]
    ) -> Generator[PolarGeometry, None, None]:
        """Internal helper for querying for clipped geometries.

        Args:
            clipping_area: The area to clip to
            only_location_type: The type of the read location

        Yields:
            The geometries clipped to the given area.
        """
        yield from self._read_geometries(
            geometry_column=f"Intersection(geometry, {clipping_area})",
            only_location_type=only_location_type,
        )

    def _read_geometries(
        self, geometry_column: str, only_location_type: Optional[LocationType]
    ) -> Generator[PolarGeometry, None, None]:
        """Factors out the common parts of assembling SQL statements for the ``read_*`` methods.

        Args:
            geometry_column:
                A SQL "column name" that can be used in a SELECT clause and which returns a Spatialite
                geometry. Examples are ``"geometry"`` to simply return the geometries unmodified or something
                like ``"Reverse(geometry) as ignored_column_name"`` to perform some modification.
                The name does not matter.
            only_location_type: Get only obstacles of that type, if not set to ``None``

        Yields:
            The geometries as read from the database.
        """
        if only_location_type is None:
            additional_where_constraint = ""
        else:
            additional_where_constraint = f"AND location_type = {only_location_type.value}"

        # `IsValid(wrapped_geometry)` excludes empty geometries which can sometimes occur
        yield from self._read_from_sql(
            f"""
            WITH temptable AS (
                SELECT id, location_type, name, ({geometry_column}) as wrapped_geometry
                FROM Obstacles
                WHERE wrapped_geometry IS NOT NULL AND IsValid(wrapped_geometry)
                {additional_where_constraint}
            )
            SELECT id, location_type, name, AsBinary(wrapped_geometry) as geometry
            FROM temptable
            """
        )

    def _read_from_sql(self, sql_statement: str) -> Generator[PolarGeometry, None, None]:  # noqa: C901
        """Reads geometries for a given complete SQL query.

        Supports reading these geometry types and maps them to instances of
        :attr:`pyrate.plan.geometry.PolarGeometry`:

            - ``Point``
            - ``LineString`` and ``MultiLineString``
            - ``Polygon`` and ``MultiPolygon``

        Args:
            sql_statement: The SQL statement to query with and read geometries from

        Yields:
            The geometries as read from the database.
        """
        with closing(self._connection.execute(sql_statement)) as cursor:
            # This should be theoretically parallelizable, but was not required as of now
            # keep in mind that `cursor.fetchall()` returns a list, not a generator
            for (identifier, location_type, name, binary_geometry) in cursor.fetchall():
                parsed_geometry = shapely.wkb.loads(binary_geometry)
                geometry_type = parsed_geometry.type

                # The database contains only the geometry types Point, LineString and Polygon.
                # However, depending on the performed operation, some entries might be cut into
                # MultiLineString or MultiPolygon, so we need to be able to decode them too.
                # MultiPoint can currently not occur.

                def to_polygon(
                    polygon: Polygon, unique_identifier: int, name=name, location_type=location_type
                ) -> PolarPolygon:
                    locations = [PolarLocation(y, x) for (x, y) in polygon.exterior.coords]
                    return PolarPolygon(locations, LocationType(location_type), name, unique_identifier)

                def to_route(
                    line_string: LineString, unique_identifier: int, name=name, location_type=location_type
                ) -> PolarRoute:
                    locations = [PolarLocation(y, x) for (x, y) in line_string.coords]
                    return PolarRoute(locations, LocationType(location_type), name, unique_identifier)

                if geometry_type == "Point":
                    point = cast(Point, parsed_geometry)
                    yield PolarLocation(
                        latitude=point.y,
                        longitude=point.x,
                        location_type=LocationType(location_type),
                        name=name,
                        identifier=identifier,
                    )

                elif geometry_type == "LineString":
                    yield to_route(cast(LineString, parsed_geometry), identifier)

                elif geometry_type == "MultiLineString":
                    for index, route in enumerate(parsed_geometry.geoms):
                        # Make identifier unique by best effort
                        yield to_route(cast(LineString, route), unique_identifier=identifier + index)

                elif geometry_type == "Polygon":
                    yield to_polygon(cast(Polygon, parsed_geometry), identifier)

                elif geometry_type == "MultiPolygon":
                    for index, polygon in enumerate(parsed_geometry.geoms):
                        # Make identifier unique by best effort
                        yield to_polygon(cast(Polygon, polygon), unique_identifier=identifier + index)

                else:  # pragma: no cover
                    # This should never happen in a well-formatted database
                    raise RuntimeError(f'illegal geometry type "{geometry_type}" returned')

    def clear(self) -> None:
        """Deletes all obstacles from the database, but does not touch the table structure or indices."""

        with self._connection:  # auto-commits at the end
            self._connection.execute("DELETE FROM Obstacles").close()

    def count_geometries(self) -> int:
        """Counts all obstacles in the database."""

        with closing(self._connection.execute("SELECT COUNT(*) FROM Obstacles")) as cursor:
            result = cursor.fetchone()

        return cast(int, result[0])  # needs to be cast as the result is untyped

    def count_vertices(self) -> int:
        """Counts all vertices of all obstacles in the database."""

        statement = "SELECT SUM(ST_NPoints(geometry)) FROM Obstacles"
        with closing(self._connection.execute(statement)) as cursor:
            result = cursor.fetchone()

        count = cast(Optional[int], result[0])  # needs to be cast as the result is untyped
        if count is None:
            # this can happen if the database is empty since `SUM` will return `NULL` in that case
            return 0

        return count

    def __len__(self) -> int:
        return self.count_geometries()

    def close(self) -> None:
        """Closes the connection to the database and releases all associated resources.

        It is not really documented in the standard library but :meth:`sqlite3.Connection.close` and thus
        this method can apparently be called multiple times.
        """

        # The `_connection` is unset if an exception has been thrown, but
        # `close()` might still be called when the database was used as a
        # context manager
        if hasattr(self, "_connection"):
            self._connection.close()  # pragma: no cover

    def __enter__(self) -> "SpatialiteDatabase":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    # Makes sure the connection is closed when this object ceases to exist
    def __del__(self) -> None:
        self.close()


def _random_name() -> str:
    """Returns a probably unique random name consisting only of latin letters."""
    return "".join(random.choices(string.ascii_letters, k=32))


def to_wkb(geometry: PolarGeometry) -> bytes:
    """Converts the given geometries into well-known binary (WKB) bytes.

    Args:
        geometry: The polar geometry to be converted

    Returns:
        The WKB representation of the geometry

    Raises:
        NotImplementedError:
            If the geometry type cannot be converted to bytes.
            This will never occur when the type signature is obeyed.
    """

    if isinstance(geometry, PolarLocation):
        return cast(bytes, shapely.wkb.dumps(Point(geometry.longitude, geometry.latitude)))
    if isinstance(geometry, PolarRoute):
        return cast(bytes, shapely.wkb.dumps(LineString(geometry.to_numpy())))
    if isinstance(geometry, PolarPolygon):
        return cast(bytes, shapely.wkb.dumps(Polygon(geometry.to_numpy())))

    # Can never occur if the type signature was obeyed but better be explicit here
    raise NotImplementedError(f"unknown geometry type: {type(PolarLocation).__name__}")
