"""The module contains methods to access raster data sets (as opposed to vector data sets).

The :class:`~pyrate.common.raster_datasets.geo_datasets.DataSetAccess` allows to read data arrays from
raster datasets using query windows. It also computes such windows for a given point and radius.
However, client code will often want to use some transformed properties of these datasets.
To that end, a concrete :class:`~pyrate.common.raster_datasets.transformer_base.BaseTransformer` can be used,
either implemented in the :mod:`~pyrate.common.raster_datasets.transformers_concrete` module
or in some client code.
Transformers query some data source for given nodes and radii of influence and then return some
- well, transformed - property vectors for the query nodes.
By the way: Instances of :class:`~pyrate.plan.graph.geo_graph.GeoNavigationGraph`
directly accept instances of :class:`~pyrate.common.raster_datasets.transformer_base.BaseTransformer`
to generate properties for nodes in the graph.

.. inheritance-diagram::
        pyrate.common.raster_datasets.transformers_concrete.ConstantTransformer
        pyrate.common.raster_datasets.transformers_concrete.BathymetricTransformer
    :parts: 2
    :top-classes: pyrate.common.raster_datasets.transformer_base.BaseTransformer

You might want to set some options for the underlying *rasterio*/*GDAL* drivers, like
`GDAL_CACHEMAX <https://trac.osgeo.org/gdal/wiki/ConfigOptions#GDAL_CACHEMAX>`_
("If its value is small (less than 100000), it is assumed to be measured in megabytes, otherwise in bytes."):

.. code-block:: python

    with rasterio.Env(GDAL_CACHEMAX=1024):
        with DataSetAccess(...) as data_set:
            pass  # do cool stuff

"""

from .geo_datasets import DataSetAccess
from .transformer_base import BaseDatasetTransformer
from .transformer_base import BaseTransformer

# Don't directly expose transformers_concrete here to keep it simple

__all__ = ["BaseTransformer", "BaseDatasetTransformer", "DataSetAccess"]
