"""
This module provides the base classes for the transformers, which extract property vectors for given query
points form (usually) a geographical dataset.
"""

# Standard library
from abc import ABC
from abc import abstractmethod
from itertools import repeat

# Typing
from typing import Any
from typing import cast
from typing import ContextManager
from typing import Iterable
from typing import Sequence
from typing import Tuple
from typing import Union

# Scientific
from numpy import array
from numpy import empty
from numpy import ndarray
from numpy.typing import DTypeLike
from pandas import DataFrame

# Progress bar
from tqdm import tqdm

# Typing helpers
from .geo_datasets import DataSetAccess


class BaseTransformer(ContextManager["BaseTransformer"], ABC):
    """This class allows to query some data source for a property at each node.

    Subclasses will usually override :meth:`_get_transformed_at` in order to return the data vector for some
    specific node with given latitude and longitude. Note that the result of calling
    :meth:`~get_transformed_at_nodes` is a :class:`pandas.DataFrame`, in order to allow a single transformer
    to return multiple values for each vector if this simplifies or speeds up calculations.

    Querying all nodes prints a progress bar to the command line by default.

    Warning:
        This class (and any subclasses) shall only be used as a context manager. See :class:`~DataSetAccess`
        for the reasons for this.

    Args:
        structured_dtype: For each column in the query result a tuple consisting of a human-readable names and
            the (numpy) data type of property. This follows the syntax of NumPy's
            `"Structured Datatypes" <https://numpy.org/doc/stable/user/basics.rec.html#structured-datatypes>`_
            .

    See Also:
        BaseDatasetTransformer
    """

    def __init__(self, structured_dtype: Sequence[Tuple[str, DTypeLike]]) -> None:
        super().__init__()
        self.structured_dtype = cast(DTypeLike, structured_dtype)

    @abstractmethod
    def _get_transformed_at(self, latitude: float, longitude: float, radius: float) -> Tuple:
        """Get the property at some specific node, given by its geographical location.

        Args:
            latitude: The geographical location of the node, in radians
            longitude: The geographical location of the node, in radians
            radius: The radius of the area that this node shall represent, in meters

        Returns:
            A single property vector for each single node.
        """

    def get_transformed_at_nodes(
        self,
        latitudes: ndarray,
        longitudes: ndarray,
        radius: Union[float, ndarray],
        show_progress: bool = False,
    ) -> DataFrame:
        """Computes the property for each individual node. Prints a progress bar by default.

        Args:
            latitudes: latitude values of all nodes, in radians
            longitudes: longitude values of all nodes, in radians
            radius: the radius around each node that it should represent, in meters; may be an array of shape
                    `(num_nodes, )` or a single scalar if the radius is uniform
            show_progress: whether to print a nice and simple progress bar

        Returns:
            An array of all values generated for each (latitude, longitude) node, with shape
            `(number of nodes, number of properties per node)`
        """

        assert latitudes.shape == longitudes.shape

        radii = repeat(radius) if isinstance(radius, float) else cast(Iterable[float], radius)

        if len(latitudes) > 0:
            result = [
                self._get_transformed_at(latitude, longitude, rad)
                for latitude, longitude, rad in tqdm(
                    zip(latitudes, longitudes, radii),
                    unit=" nodes",
                    unit_scale=True,
                    colour="white",
                    total=len(latitudes),
                    disable=not show_progress,
                )
            ]
        else:
            result = empty((0, len(self.structured_dtype)))  # type: ignore

        assert len(result) == latitudes.shape[0]

        # this also ensures that all property vectors have the same length
        structured_array = array(result, dtype=self.structured_dtype)
        return DataFrame.from_records(structured_array)

    def __enter__(self) -> "BaseTransformer":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Any:
        return False


class BaseDatasetTransformer(BaseTransformer, ABC):
    """A specialized dataset transformer which extracts properties from a geographical dataset.

    Args:
        structured_dtype: see constructor argument ``structured_dtype`` in :class:`BaseTransformer`
        dataset: A dataset to read from.
                 It is automatically managed when this class is used as a context manager.
    """

    def __init__(self, structured_dtype: Sequence[Tuple[str, DTypeLike]], dataset: DataSetAccess) -> None:
        super().__init__(structured_dtype)
        self.dataset = dataset

    def __enter__(self) -> "BaseDatasetTransformer":
        self.dataset.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Any:
        super().__exit__(exc_type, exc_val, exc_tb)
        return self.dataset.__exit__(exc_type, exc_val, exc_tb)
