"""
This module exposes specific property transformers.

See the `data repository <https://gitlab.sailingteam.hg.tu-darmstadt.de/informatik/data>`_ for details on
the actual data sets referenced and used here.
"""

# Standard library
from enum import auto
from enum import Enum

# Typing
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

# Scientific
import numpy
from numpy import clip
from numpy import count_nonzero
from numpy import extract
from numpy import float32
from numpy.typing import DTypeLike

# Helpers and own typing
from ...plan.geometry.helpers import haversine_numpy
from .geo_datasets import DataSetAccess
from .transformer_base import BaseDatasetTransformer
from .transformer_base import BaseTransformer


class ConstantTransformer(BaseTransformer):

    """A very simple transformer class to fill a property with a constant value.

    Args:
        value: The constant value to be added to each node
        dtype: The numpy data type of the resulting property vector field
        name: The name of the property. If set to ``None``, a reasonable default name will be used.
    """

    # pylint: disable=too-few-public-methods

    # idea: could make this class explicitly generic in the type signatures
    def __init__(self, value: Any, dtype: DTypeLike, name: Optional[str] = None) -> None:
        name = f"constant value of {value}" if name is None else name
        super().__init__([(name, dtype)])
        self.value = value

    def _get_transformed_at(self, latitude: float, longitude: float, radius: float) -> Tuple[Any]:
        return (self.value,)  # return a tuple


class BathymetricTransformer(BaseDatasetTransformer):

    """Extracts values from a given bathymetric datasets, i.e. depth information.

    The datatype for all modes is ``np.float32``. Raises a :class:`ValueError` if not data is found for any
    given query ``(latitude, longitude, radius)``.

    Args:
        dataset: the data set to be used
        modes: a sequence of modes of the values to be extracted; see :class:`~BathymetricTransformer.Modes`
    """

    # pylint: disable=too-few-public-methods

    #: The depth/elevation below which a point is considered navigable, in meters;
    #: positive values are above sea level, negative ones below
    NAVIGABLE_BELOW: float = -5.0

    class Modes(Enum):
        """The different modes how depth values can be extracted."""

        AVERAGE_DEPTH = auto()
        """The average depth, weighted with more emphasis on the area near the query point, in meters.

        This is a visualisation of the mode when applied to the `Earth 2014
        <https://gitlab.sailingteam.hg.tu-darmstadt.de/informatik/data/-/tree/master/topography/earth2014>`_
        topographic/bathymetric dataset:

        .. image:: plot_global_bathymetry_depth.png
          :alt: AVERAGE_DEPTH mode when applied to the Earth 2014 topographic/bathymetric dataset
        """

        FRACTION_NAVIGABLE = auto()
        """The fraction of data points at which a boat can navigate, as a scalar in :math:`[0, 1]`.

        See :attr:`BathymetricTransformer.NAVIGABLE_BELOW` for what is considered navigable.
        Only the topographic/bathymetric height/depth value is used for determining navigability as an
        approximation, and not actual water coverage.
        To not count the Netherlands as navigable, the value is set to a vale a little bit below zero,
        where zero means sea level.

        This is a visualisation of the mode when applied to the `Earth 2014
        <https://gitlab.sailingteam.hg.tu-darmstadt.de/informatik/data/-/tree/master/topography/earth2014>`_
        topographic/bathymetric dataset:

        .. image:: plot_global_bathymetry_fraction_navigable.png
          :alt: FRACTION_NAVIGABLE mode when applied to the Earth 2014 topographic/bathymetric dataset
        """

        @property
        def column_name(self) -> str:
            """Returns the name that is used for the dataframe column."""
            return f"bathymetric data ({self.name})"

    def __init__(self, dataset: DataSetAccess, modes: Sequence[Modes] = tuple(Modes)) -> None:
        structured_dtype = [(mode.column_name, float32) for mode in modes]
        super().__init__(structured_dtype, dataset)
        self.modes = modes

    def _get_transformed_at(self, latitude: float, longitude: float, radius: float) -> Tuple:
        # pylint: disable=too-many-locals
        assert radius > 0, "the radius must be positive"

        windows = self.dataset.get_bounding_windows_around(
            center_latitude=latitude, center_longitude=longitude, radius=radius
        )
        lats, lons = self.dataset.lat_lon_meshgrid_for(*windows, radians=True)
        assert lats.shape == lons.shape

        # the depth data is the negative for below-sea level and positive for landmasses
        depth = self.dataset.data_for(*windows)  # as int16 !
        assert depth.shape == lats.shape

        # get the distances to all points
        distances = haversine_numpy(lats, lons, latitude, longitude)

        result: List[float] = []  # will be transformed to float32 later

        for mode in self.modes:  # respect the ordering

            if mode is BathymetricTransformer.Modes.AVERAGE_DEPTH:
                # we simply use the inverse distance as a weight for the weighted arithmetic mean
                weights = clip(1.0 - (distances / radius), 0.0, 1.0)
                try:
                    average = numpy.average(depth.astype(float32), weights=weights)
                except ZeroDivisionError as error:
                    raise ValueError(
                        f"no points in radius {radius} m around (lat, lon) = {(latitude, longitude)} rad"
                    ) from error
                result.append(average)

            elif mode is BathymetricTransformer.Modes.FRACTION_NAVIGABLE:
                # we use the distance to cut off unwanted entries
                depth_within_radius = extract(distances <= radius, depth)  # also flattens
                if len(depth_within_radius) == 0:
                    raise ValueError(
                        f"no points in radius {radius} m around (lat, lon) = {(latitude, longitude)} radians"
                    )
                number_of_navigable = count_nonzero(
                    depth_within_radius <= BathymetricTransformer.NAVIGABLE_BELOW
                )
                fraction = number_of_navigable / len(depth_within_radius)
                result.append(fraction)

            else:  # pragma: no branch
                raise ValueError(f"invalid mode {mode}")  # pragma: no cover

        return tuple(result)
