"""This module contains helpers for writing and running tests.

In particular, it contains flags for detecting specific scenarios (like the test running on a CI platform)
and a variety of custom strategies for *Hypothesis*.
"""

# Standard library
from os import environ

# Spatialite for environment setup
from pyrate.common.charts import SpatialiteDatabase


def env(name: str) -> bool:
    """Checks if an environment variable exists and its value in lower case is ``{yes, true, t, 1}``.

    Args:
        name: The name of the environment variable to check
    """

    return environ.get(name, "").lower() in ("yes", "true", "t", "1")


#: Set to ``True`` if running on a CI server (best effort)
IS_CI: bool = env("CI") or env("CONTINUOUS_INTEGRATION")

#: Whether to intensify tests at the expense of more time, e.g. with more example data for hypothesis
IS_EXTENDED_TESTING: bool = env("EXTENDED_TESTING")


#: True iff the Spatialite SQLite extension is installed & can be used
SPATIALITE_AVAILABLE: bool
try:
    with SpatialiteDatabase(":memory:"):
        pass  # make sure it is properly closed
except RuntimeError:  # pragma: no cover
    SPATIALITE_AVAILABLE = False
else:
    SPATIALITE_AVAILABLE = True
