"""Contains helpers based on hypothesis test data generators."""

# Typing
from typing import cast
from typing import Tuple

# Hypothesis testing
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import composite
from hypothesis.strategies import floats
from hypothesis.strategies import lists
from hypothesis.strategies import SearchStrategy

# Mathematics
from numpy import eye
from numpy import float64

# Gaussian representation
from pyrate.common.math import Gaussian

# Own typing
from . import DrawType  # pylint: disable=unused-import


# In this context, we reproduce a common filter notation
# pylint: disable=invalid-name, too-many-locals, unused-argument


@composite
def linear_model(
    draw: DrawType, state_dim: int = 2, input_dim: int = 1, sensor_dim: int = 2
) -> SearchStrategy[Tuple]:
    """Generate a linear state space model.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        state_dim: Number of state variables
        input_dim: Number of input variables
        sensor_dim: Number of measurement variables
    """

    # Strategies
    float_strategy = floats(0, 5)

    # Transition model
    F = draw(arrays(float64, (state_dim, state_dim), elements=float_strategy))

    # Input model
    B = draw(arrays(float64, (state_dim, input_dim), elements=float_strategy))

    # Measurement model
    H = draw(arrays(float64, (sensor_dim, state_dim), elements=float_strategy))

    # Symmetric, positive definite process noise
    q = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    Q = q @ q.T + eye(state_dim)

    # Symmetric, positive definite sensor noise
    r = draw(arrays(float64, (sensor_dim, 1), elements=float_strategy))
    R = r @ r.T + eye(sensor_dim)

    # Initial belief
    x = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    p = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    P = p @ p.T + eye(state_dim)

    estimate = Gaussian(x, P)

    # Measurements and inputs
    measurements = draw(
        lists(arrays(float64, (sensor_dim, 1), elements=float_strategy), min_size=2, max_size=4)
    )
    inputs = draw(
        lists(
            arrays(float64, (input_dim, 1), elements=float_strategy),
            min_size=len(measurements),
            max_size=len(measurements),
        )
    )

    # Return model
    result = estimate, F, B, H, Q, R, measurements, inputs
    return cast(SearchStrategy[Tuple], result)


@composite
def nonlinear_model(draw: DrawType, state_dim: int = 2, sensor_dim: int = 2) -> SearchStrategy[Tuple]:
    """Generate a nonlinear state space model.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        state_dim: Number of state variables
        sensor_dim: Number of measurement variables
    """

    # Strategies
    float_strategy = floats(0, 5)

    # Transition model
    F = draw(arrays(float64, (state_dim, state_dim), elements=float_strategy))

    def f(x):
        return F @ x

    # Jacobi of f about state
    def Jf(x):
        return F

    # Measurement model
    H = draw(arrays(float64, (sensor_dim, state_dim), elements=float_strategy))

    def h(x):
        return H @ x

    # Jacobi of h about state
    def Jh(state):
        return H

    # Symmetric, positive definite process noise
    q = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    Q = q @ q.T + eye(state_dim)

    # Symmetric, positive definite sensor noise
    r = draw(arrays(float64, (sensor_dim, 1), elements=float_strategy))
    R = r @ r.T + eye(sensor_dim)

    # Initial belief
    x = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    p = draw(arrays(float64, (state_dim, 1), elements=float_strategy))
    P = p @ p.T + eye(state_dim)

    estimate = Gaussian(x, P)

    # Measurements
    measurements = draw(
        lists(arrays(float64, (sensor_dim, 1), elements=float_strategy), min_size=2, max_size=4)
    )

    # Return model
    result = estimate, f, F, Jf, h, H, Jh, Q, R, measurements
    return cast(SearchStrategy[Tuple], result)
