"""Contains helpers like hypothesis test data generators."""

# Typing
from typing import Any
from typing import cast
from typing import Dict
from typing import Optional
from typing import Union

# Hypothesis testing
from hypothesis import assume
import hypothesis.extra.numpy as numpy_st
import hypothesis.strategies as st
from hypothesis.strategies import SearchStrategy

# Scientific stack
import numpy
from scipy.spatial import Voronoi

# Planning primitives
from shapely.geometry import box
from shapely.geometry import MultiLineString
from shapely.geometry import Polygon
from shapely.ops import polygonize
from shapely.ops import unary_union

# Geospatial objects
from pyrate.plan.geometry import CartesianGeometry
from pyrate.plan.geometry import CartesianLocation
from pyrate.plan.geometry import CartesianPolygon
from pyrate.plan.geometry import CartesianRoute
from pyrate.plan.geometry import Geospatial
from pyrate.plan.geometry import LocationType
from pyrate.plan.geometry import PolarGeometry
from pyrate.plan.geometry import PolarLocation
from pyrate.plan.geometry import PolarPolygon
from pyrate.plan.geometry import PolarRoute

# Own typing
from . import DrawType  # pylint: disable=unused-import


@st.composite
def geo_bearings(draw: DrawType) -> SearchStrategy[float]:
    """Returns a direction/bearing/azimuth/yaw for navigation in degrees in :math:`[0, 360)`.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    bearing = draw(st.floats(min_value=0.0, max_value=360.0, exclude_max=True))
    return cast(SearchStrategy[float], bearing)


@st.composite
def geospatial_identifiers(draw: DrawType) -> SearchStrategy[Optional[int]]:
    """Returns identifiers for subclasses of :class:`pyrate.plan.geometry.Geospatial`.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    identifier = draw(st.one_of(st.none(), st.integers(min_value=0, max_value=(2**63) - 1)))
    return cast(SearchStrategy[Optional[int]], identifier)


@st.composite
def location_types(draw: DrawType) -> SearchStrategy[LocationType]:
    """Returns location types.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    location_type = draw(st.sampled_from(LocationType))
    return cast(SearchStrategy[LocationType], location_type)


@st.composite
def geospatial_attributes(draw: DrawType) -> SearchStrategy[Dict[str, Any]]:
    """Returns the common attributes for subclasses of :class:`pyrate.plan.geometry.Geospatial`.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    attributes = {
        "location_type": draw(location_types()),
        "name": draw(st.text()),
        "identifier": draw(geospatial_identifiers()),
    }
    return cast(SearchStrategy[Dict[str, Any]], attributes)


@st.composite
def polar_locations(draw: DrawType) -> SearchStrategy[PolarLocation]:
    """Returns a polar location.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    location = PolarLocation(
        draw(st.floats(min_value=-90.0, max_value=+90.0)),
        draw(st.floats(min_value=-180.0, max_value=+180.0, exclude_max=True)),
        **draw(geospatial_attributes()),
    )
    return cast(SearchStrategy[PolarLocation], location)


@st.composite
def cartesian_locations(
    draw: DrawType,
    origin: SearchStrategy[Union[CartesianPolygon, None]] = st.one_of(st.none(), polar_locations()),
) -> SearchStrategy[CartesianLocation]:
    """Returns a cartesian location.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        origin: an optional strategy for specifying origins, defaults to providing both ``None`` and real
                locations
    """
    location = CartesianLocation(
        draw(st.floats(min_value=-10_000.0, max_value=+10_000.0)),
        draw(st.floats(min_value=-10_000.0, max_value=+10_000.0)),
        origin=draw(origin),
        **draw(geospatial_attributes()),
    )
    return cast(SearchStrategy[CartesianLocation], location)


@st.composite
def polar_objects(
    draw: DrawType, stable: bool = False, non_repeating: bool = False
) -> SearchStrategy[PolarGeometry]:
    """Returns polar geometries.

    The concrete type is sampled randomly from all three polar geometries.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        stable: see :func:`~polar_routes_stable`
        non_repeating: if ``True``, the strategy will not produce routes with duplicate locations.
                       Ignored if ``stable`` is given.
    """
    possible_sources = st.one_of(
        [
            polar_locations(),
            polar_routes_stable() if stable else polar_routes(non_repeating=non_repeating),
            polar_polygons(),
        ]
    )
    geospatial: PolarGeometry = draw(possible_sources)
    return cast(SearchStrategy[PolarGeometry], geospatial)


@st.composite
def cartesian_objects(draw: DrawType) -> SearchStrategy[CartesianGeometry]:
    """Returns cartesian geometries.

    The concrete type is sampled randomly from all three cartesian geometries.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
    """
    possible_sources = st.one_of(
        [
            cartesian_locations(),
            cartesian_routes(),
            cartesian_polygons(),
        ]
    )
    geospatial: CartesianGeometry = draw(possible_sources)
    return cast(SearchStrategy[CartesianGeometry], geospatial)


@st.composite
def geospatial_objects(draw: DrawType, stable: bool = False) -> SearchStrategy[Geospatial]:
    """Returns instances of the abstract class :class:`pyrate.plan.geometry.Geospatial`.

    The concrete type is sampled randomly from all six cartesian and polar geometries.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        stable: see :func:`~polar_routes_stable`
    """
    geospatial: Geospatial = draw(st.one_of([polar_objects(stable=stable), cartesian_objects()]))
    return cast(SearchStrategy[Geospatial], geospatial)


@st.composite
def cartesian_polygons(  # pylint: disable=too-many-arguments,too-many-locals
    draw: DrawType,
    min_vertices: int = 5,
    max_vertices: int = 15,
    scale: float = 100_000,
    center_x: float = 0.0,
    center_y: float = 0.0,
    origin: SearchStrategy[Union[CartesianPolygon, None]] = st.one_of(st.none(), polar_locations()),
) -> SearchStrategy[CartesianPolygon]:
    """Returns non-empty valid cartesian polygons around the origin of the coordinate system.

    Inspired `by testing code from the spatialpandas
    <https://github.com/holoviz/spatialpandas/blob/efdabe5c736db8103a4bfedca55a414a365b754a/tests/geometry/strategies.py#L141>`_
    library.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        min_vertices: the minimum number of locations that shall form this polygon; needs to be at least ``5``
                      such that the generation algorithm works reliably
        max_vertices: the minimum number of locations that shall form this polygon; needs to be larger than
                      ``min_vertices``; this may not be very large as this will make example generation
                      extremely slow
        scale: the maximum that a single coordinate value may be away from the center (in meters)
        center_x: the east-west center (in meters)
        center_y: the north-south center (in meters)
        origin: an optional strategy for specifying origins, defaults to providing both ``None`` and real
                locations

    Raises:
        ValueError: if polygon generation fails
    """

    assert min_vertices >= 5, "min_vertices needs to be at least 5"
    assert max_vertices >= min_vertices, "max_vertices needs to be at least min_vertices"
    assert scale >= 0.0, "scale must be non-negative"

    count = draw(st.integers(min_value=min_vertices, max_value=max_vertices))

    # very often, this only takes a single try
    # it is highly unlikely that it will take more than 50
    tries = 50
    for _ in range(tries):  # pragma: no branch

        # create points in [-0.5, +0.5]
        points = numpy.random.rand(count, 2) - 0.5
        # scale them to the desired size
        points *= scale * 2

        voronoi = Voronoi(points)
        multi_line_string = MultiLineString(
            [voronoi.vertices[s] for s in voronoi.ridge_vertices if all(numpy.array(s) >= 0)]
        )

        poly = unary_union(list(polygonize(multi_line_string)))
        poly = poly.intersection(box(-scale, -scale, scale, scale))

        if (  # pragma: no branch
            isinstance(poly, Polygon) and not poly.is_empty and poly.is_simple and poly.is_valid
        ):
            coordinates = numpy.array(poly.exterior.coords)
            # move them to the desired center
            coordinates[:, 0] += center_x
            coordinates[:, 1] += center_y

            polygon = CartesianPolygon.from_numpy(
                coordinates, origin=draw(origin), **draw(geospatial_attributes())
            )
            return cast(SearchStrategy[CartesianPolygon], polygon)

    # This should practically never occur (the probability is very, very low)
    raise ValueError("Failed to construct polygon")  # pragma: no cover


@st.composite
def polar_polygons(
    draw: DrawType,
    min_vertices: int = 5,
    max_vertices: int = 15,
    scale: float = 100_000,
    center: Optional[PolarLocation] = None,
) -> SearchStrategy[PolarPolygon]:
    """Returns non-empty valid polar polygons.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        min_vertices: the minimum number of locations that shall form this polygon; needs to be at least ``5``
                      such that the generation algorithm works reliably
        max_vertices: the minimum number of locations that shall form this polygon; needs to be larger than
                      ``min_vertices``; this may not be very large as this will make example generation
                      extremely slow
        scale: the maximum that a single coordinate value may be away from the center (in meters)
        center: the center of the polygon or ``None`` randomly select one

    Raises:
        ValueError: if polygon generation fails
    """
    cartesian = draw(
        cartesian_polygons(
            min_vertices=min_vertices,
            max_vertices=max_vertices,
            scale=scale,
            origin=polar_locations() if center is None else st.just(center),  # type: ignore
        )
    )
    return cast(SearchStrategy[PolarPolygon], cartesian.to_polar())


@st.composite
def cartesian_routes(  # pylint: disable=too-many-arguments
    draw: DrawType,
    min_vertices: int = 2,
    max_vertices: int = 10,
    scale: float = 100_000,
    center_x: float = 0.0,
    center_y: float = 0.0,
    origin: SearchStrategy[Union[CartesianPolygon, None]] = st.one_of(st.none(), polar_locations()),
    non_repeating: bool = True,
) -> SearchStrategy[CartesianRoute]:
    """Returns a cartesian route.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        min_vertices: the minimum number of locations that shall form this route, must be ``2`` or greater
        max_vertices: the maximum number of locations that shall form this route
        scale: the maximum that a single coordinate value may be away from the center (in meters); strictly
            positive
        center_x: the east-west center (in meters)
        center_y: the north-south center (in meters)
        origin: an optional strategy for specifying origins, defaults to providing both ``None`` and real
            locations
        non_repeating: if ``True``, the route will not contain any duplicate locations
    """

    assert min_vertices >= 2, "min_vertices may not be less than 2"
    assert (
        max_vertices is None or max_vertices >= min_vertices
    ), "max_vertices may not be less than min_vertices"
    assert scale > 0.0, "scale must be strictly positive"

    # define the actual values in the coordinate arrays
    elements_x = st.floats(min_value=center_x - scale, max_value=center_x + scale)
    elements_y = st.floats(min_value=center_y - scale, max_value=center_y + scale)

    # define the number of coordinates; we must draw directly and not pass the strategy such that x and y have
    # the same number of elements
    length = draw(st.integers(min_value=min_vertices, max_value=max_vertices))

    # create the actual coordinates and ensure that all are different from each other
    coordinates_x = draw(numpy_st.arrays(dtype="float64", shape=length, elements=elements_x, unique=True))
    coordinates_y = draw(numpy_st.arrays(dtype="float64", shape=length, elements=elements_y, unique=True))
    coordinates = numpy.vstack((coordinates_x, coordinates_y)).T

    # make sure that the route has non-zero length
    if numpy.abs(numpy.diff(coordinates, axis=1)).sum() < 1:  # one meter in the typical interpretation
        coordinates[0, 0] += 1  # add an arbitrary value  # pragma: no cover

    assume(not non_repeating or (numpy.abs(numpy.diff(coordinates, axis=1)) > 1).all())  # Difficult to handle

    # create the route with the other parameters of geospatial objects
    route = CartesianRoute.from_numpy(coordinates, origin=draw(origin), **draw(geospatial_attributes()))
    return cast(SearchStrategy[CartesianRoute], route)


@st.composite
def polar_routes(
    draw: DrawType,
    min_vertices: int = 2,
    max_vertices: int = 10,
    non_repeating: bool = True,
) -> SearchStrategy[PolarRoute]:
    """Returns a polar route.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        min_vertices: the minimum number of locations that shall form this route, must be ``2`` or greater
        max_vertices: the maximum number of locations that shall form this route or ``None`` to let
            *hypothesis* decide
        non_repeating: if ``True``, the route will not contain any duplicate locations
    """

    assert min_vertices >= 2, "min_vertices may not be less than 2"
    assert (
        max_vertices is None or max_vertices >= min_vertices
    ), "max_vertices may not be less than min_vertices"

    # define the actual values in the coordinate arrays
    elements_latitude = st.floats(min_value=-90, max_value=90)
    elements_longitude = st.floats(min_value=-180, max_value=180, exclude_max=True)

    # define the number of coordinates; we must draw directly and not pass the strategy such that x and y have
    # the same number of elements
    length = draw(st.integers(min_value=min_vertices, max_value=max_vertices))

    # create the actual coordinates
    coordinates_latitude = draw(
        numpy_st.arrays(dtype="float64", shape=length, elements=elements_latitude, unique=True)
    )
    coordinates_longitude = draw(
        numpy_st.arrays(dtype="float64", shape=length, elements=elements_longitude, unique=True)
    )
    coordinates = numpy.vstack((coordinates_latitude, coordinates_longitude)).T

    # make sure that the route has non-zero length
    # there is no single correct value for the threshold near the poles, but 1e-4 appears to work fine
    if numpy.abs(numpy.diff(coordinates, axis=1)).sum() < 1e-4:
        coordinates[0, 0] += 1e-4  # add an arbitrary value  # pragma: no cover

    assume(
        not non_repeating or (numpy.abs(numpy.diff(coordinates, axis=1)) > 1e-4).all()
    )  # Difficult to handle

    # create the route with the other parameters of geospatial objects
    try:
        route = PolarRoute.from_numpy(coordinates, **draw(geospatial_attributes()))
    except ValueError:
        # This can still happen if the duplicate entries check above does not catch it
        assume(False)  # pragma: no cover

    # Make sure we only generate routes that can be projected
    try:
        route.to_cartesian(route.locations[0])
    except AssertionError:
        assume(False)  # pragma: no cover

    return cast(SearchStrategy[PolarRoute], route)


@st.composite
def polar_routes_stable(
    draw: DrawType,
    min_vertices: int = 5,  # polar_polygons() requires at least 5
    max_vertices: int = 10,
) -> SearchStrategy[PolarRoute]:
    """Returns a polar route where the vertices are not too far apart.

    It is therefore numerically more stable when projecting to a cartesian plane.

    Args:
        draw: see :func:`hypothesis.strategies.composite`
        min_vertices: the minimum number of locations that shall form this route, must be ``5`` or greater
        max_vertices: the maximum number of locations that shall form this route or ``None`` to let
            *hypothesis* decide
    """

    assert min_vertices >= 5, "min_vertices may not be less than 5"
    assert (
        max_vertices is None or max_vertices >= min_vertices
    ), "max_vertices may not be less than min_vertices"

    # We create a polygon since that is known to cause fewer numerical issues when projecting
    # since we generate them in a way that guarantees similar locality of the vertices
    polygon: PolarPolygon = draw(polar_polygons(min_vertices=min_vertices, max_vertices=max_vertices))

    return cast(
        SearchStrategy[PolarRoute],
        PolarRoute.from_numpy(
            polygon.to_numpy(),
            location_type=polygon.location_type,
            name=polygon.name,
            identifier=polygon.identifier,
        ),
    )
