"""This module implements abstractions for timestamped geospatial locations in WGS84 and local coordinates.

Two locations are ``==`` if and only if they are equal according to ``equals_exact()``.
"""

# Standard library
from copy import deepcopy
from math import cos
from math import radians
from math import sin

# Typing
from typing import Any
from typing import cast
from typing import Dict
from typing import Optional
from typing import Tuple

# Mathematics
from geopy.distance import GeodesicDistance
from geopy.distance import GreatCircleDistance
from numpy import array
from numpy import ndarray
from pyproj import Proj
from shapely.affinity import translate
from shapely.geometry import Point

# Basis
from .geospatial import Geospatial
from .geospatial import LocationType

# Helpers
from .helpers import normalize_latitude
from .helpers import normalize_longitude
from .helpers import translate_floats


class PolarLocation(Geospatial):

    """A geospatial location representing a spatial object on earth.

    See `here <http://www.movable-type.co.uk/scripts/latlong.html>`__ for a nice collection of formulas and
    explanations on geographic transformations and calculations. This is the *Rome* for geographic calculation
    questions on *Stack Overflow*: All roads seem to eventually lead here.

    Examples:
        First import some packages

        >>> from math import isclose
        >>> from pyrate.plan.geometry import PolarLocation, Direction

        Then create two example coordinates to work with:

        >>> team_room = PolarLocation(latitude=49.878091, longitude=8.654052)
        >>> frankfurt = PolarLocation(latitude=50.113709, longitude=8.656561)

        Translate the team room 27 km north, which is towards *Frankfurt*:

        >>> team_room, direction_going_back = team_room.translate(direction=Direction.North, distance=27_000)
        >>> assert isclose(direction_going_back, Direction.South)

        The variable ``team_room`` now represents a location in/near *Frankfurt*,
        only a couple hundred meters away from the location ``frankfurt``:

        >>> print(team_room.distance(frankfurt))  # doctest: +ELLIPSIS
        812.512...

        Coordinats can also be projected onto a local tanged plane and back.
        The ``origin`` defines the point where the plane touches the sphere.

        >>> frankfurt == frankfurt.to_cartesian(origin=frankfurt).to_polar()
        True

    Args:
        latitude: The latitude in degrees (will be normalized)
        longitude: The longitude in degrees (will be normalized)
        location_type: The type of this polygon
        name: An optional name of this polygon
        identifier: An optional unique identifier for this object, in :math:`[0, 2**63)`, i.e. 64 signed bits
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        latitude: float,
        longitude: float,
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
    ) -> None:
        # Type hints
        self._latitude: float
        self._longitude: float
        self._projection: Optional[Proj]

        # Attributes setup
        self.latitude = latitude
        self.longitude = longitude
        # self._projection = None  # already set by the property accesses before

        super().__init__(location_type=location_type, name=name, identifier=identifier)

    @property
    def latitude(self) -> float:
        """The latitude of this location in degrees in :math:`[-90, +90]`.

        The value is always disambiguated/normalized.
        """

        return self._latitude

    @latitude.setter
    def latitude(self, latitude: float) -> None:
        self._projection = None
        self._latitude = normalize_latitude(latitude)

    @property
    def longitude(self) -> float:
        """The longitude of this location degrees in in :math:`[-180, +180)`.

        The value is always disambiguated/normalized.
        """

        return self._longitude

    @longitude.setter
    def longitude(self, longitude: float) -> None:
        self._projection = None
        self._longitude = normalize_longitude(longitude)

    @property
    def projection(self) -> Proj:
        """Derive a :class:`pyproj.Proj` instance for projecting points.

        This instance is cached for performance reasons, since its creation is relatively time consuming. The
        cache is appropriately invalidated when setting a new :attr:`~latitude` or :attr:`~longitude`.
        """

        if self._projection is None:
            self._projection = Proj(
                proj="tmerc",
                ellps="WGS84",
                units="m",
                lon_0=self.longitude,
                lat_0=self.latitude,
            )

        return self._projection

    def to_cartesian(self, origin: "PolarLocation") -> "CartesianLocation":
        """Projects this point to a cartesian representation according to the given global reference.

        Args:
            origin: The reference by which to project onto the local tangent plane

        Returns:
            The cartesian representation of this point with the given reference point being set
        """

        # convert to cartesian
        east, north = origin.projection(self.longitude, self.latitude)

        return CartesianLocation(
            east,
            north,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
            origin=origin,
        )

    def translate(self, direction: float, distance: float) -> Tuple["PolarLocation", float]:
        """Translates this location and returns the new location and back-azimuth.

        See :func:`pyrate.plan.geometry.helpers.translate_floats` for details.
        """

        back_azimuth: float  # this is required for mypy, don't know why that is
        (longitude, latitude), back_azimuth = translate_floats(
            self.longitude, self.latitude, direction, distance
        )

        new_location = PolarLocation(
            longitude=longitude,
            latitude=latitude,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
        )

        return new_location, back_azimuth

    def distance(self, other: "PolarLocation", approximate: bool = False) -> float:
        """Calculate the horizontal geodesic distance to another location in meters, assumes degrees.

        This assumes an ellipsoidal earth and converges for any pair of points on earth.
        It is accurate to round-off and uses *geographiclib* (https://pypi.org/project/geographiclib/)
        via *geopy* (https://pypi.org/project/geopy/).

        The faster *great-circle distance* can also be used by setting *approximate=True*.
        It assumes only a spherical earth and is guaranteed to give a result for any pair of points.
        It is wrong by up to 0.5% and based on *geopy*. It is advised to use the exact solution unless you
        know what you are doing.

        See also:
            - https://en.wikipedia.org/wiki/Geodesics_on_an_ellipsoid
            - https://en.wikipedia.org/wiki/Great-circle_distance
            - https://en.wikipedia.org/wiki/Geographical_distance

        Args:
            other: The location to measure the distance to in degrees
            approximate: Whether to use a faster approximation or not (default: ``False``)

        Returns:
            The distance to the other point in meters
        """

        # input as latitude, longitude
        this = (self.latitude, self.longitude)
        that = (other.latitude, other.longitude)

        if approximate:
            distance = GreatCircleDistance(this, that).meters
        else:
            distance = GeodesicDistance(this, that).meters

        # Geopy is not typed as of now
        return cast(float, distance)

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {"type": "Point", "coordinates": (self.longitude, self.latitude)}

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    def equals(self, other: Any) -> bool:
        """Determines whether the given ``other`` object exactly equals this one.

        This function mimics :meth:`shapely.geometry.base.BaseGeometry.equals`:
        "Refers to point-set equality (or topological equality), and is equivalent to
        ``self.within(other) and self.contains(other)``."

        Args:
            other: The object to compare to

        Returns:
            Whether this and the other object are exactly equal
        """
        # The above docstring is also copied by PolarPolygon and PolarRoute

        return self.equals_exact(other, 0.0)

    def equals_exact(self, other: Any, tolerance: float) -> bool:
        """Determines whether the given ``other`` object equals this one.

        This function mimics :meth:`shapely.geometry.base.BaseGeometry.equals_exact`:
        "Refers to coordinate equality, which requires coordinates to be equal
        and in the same order for all components of a geometry."

        Args:
            other: The object to compare to
            tolerance: The absolute deviation in meters that is tolerated on the latitude and longitude values

        Returns:
            Whether this and the ``other`` object are (nearly) equal
        """
        # The above docstring is also copied by PolarPolygon and PolarRoute

        return (
            isinstance(other, PolarLocation)
            and Geospatial.__eq__(self, other)
            and self.distance(other) <= tolerance
        )

    def __repr__(self) -> str:
        # we leave out self._projection due to performance reasons and because it is redundant
        return f"PolarLocation(latitude={self.latitude}, longitude={self.longitude}{self._repr_extras})"


class CartesianLocation(Geospatial, Point):
    """A point in the cartesian plane based on local coordinates with an optional global reference.

    Examples:
        You can simply create a cartesion location like this, where coordinates are in meters:

        >>> location_a = CartesianLocation(east=10, north=-20)
        >>> location_b = CartesianLocation(east=-30, north=0)
        >>> distance = location_a.distance(location_b)
        >>> distance  # doctest: +ELLIPSIS
        44.721...

        Keep in mind that locations (like all other cartesian geomerties) are iummutable due to the underlying
        Shapely library:

        >>> location_a.x = 5.0
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

        The attributes ``east`` and ``north`` are provided as aliases for ``x`` and ``y``:

        >>> assert location_a.x == location_a.east
        >>> assert location_a.y == location_a.north

        You can also project them to a polar coordinate.
        To do this, one must only provide a reference point ``origin`` either when constructing the loction
        or when calling :meth:`~to_polar`:

        >>> reference = PolarLocation(latitude=50, longitude=30)
        >>> location_a.origin = reference
        >>> location_b.origin = reference
        >>> location_a.to_polar().distance(location_b.to_polar())  # doctest: +ELLIPSIS
        44.721...

        As any :class:`~CartesianLocation` also inherits from :class:`shapely.geometry.Point`,
        we can also use :mod:`shapely` methods
        (see `the Shapely docs <https://shapely.readthedocs.io/en/stable/manual.html>`__).
        For example, we can inflate the point using ``buffer()``.
        Mind though, that this will return a :mod:`shapely` geometry and not a :mod:`pyrate.plan.geometry`
        object.

        >>> buffered = location_a.buffer(10)
        >>> buffered.geometryType()
        'Polygon'

        Thus, we need to convert it back to a pyrate object like so (keep in mind that we now need a polygon):

        >>> from pyrate.plan.geometry.polygon import CartesianPolygon
        >>> buffered_pyrate = CartesianPolygon.from_shapely(buffered)
        >>> buffered.equals(buffered_pyrate)
        True

    Args:
        east: The easting of the location in meters
        north: The northing of the location in meters
        origin: A reference that can be used to project this cartesian representation (back)
            into a polar one
        location_type: The type of this polygon
        name: An optional name of this polygon
        identifier: An optional unique identifier for this object, in :math:`[0, 2**63)`, i.e. 64 signed bits
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        east: float,
        north: float,
        origin: Optional["PolarLocation"] = None,
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
    ) -> None:
        # Set attribute
        self.origin = origin

        # Typing hints (actually defined by shapely)
        self.x: float
        self.y: float

        # Initialize the super classes
        Point.__init__(self, east, north)
        Geospatial.__init__(self, location_type=location_type, name=name, identifier=identifier)

    #: Named access to the internal shapely point ``x``. Readonly.
    east: float = Point.x

    #: Named access to the internal shapely point ``y``. Readonly.
    north: float = Point.y

    def to_polar(self, origin: Optional["PolarLocation"] = None) -> PolarLocation:
        """Computes the polar representation of this point.

        Args:
            origin: The global reference to be used for back-projection, must be set if and only if
                :attr:`~pyrate.plan.geometry.CartesianLocation.origin` is ``None``

        Returns:
            The global, polar representation of this point
        """

        if origin is None:
            if self.origin is None:
                raise ValueError("need to give an explicit origin when the instance does not have one")
            origin = self.origin
        elif self.origin is not None:
            raise ValueError("provided an explicit origin while the instance already has one")

        # convert to cartesian
        longitude, latitude = origin.projection(self.east, self.north, inverse=True)

        return PolarLocation(
            longitude=longitude,
            latitude=latitude,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
        )

    @classmethod
    def from_shapely(cls, point: Point, *args, **kwargs) -> "CartesianLocation":
        """Create a cartesian location from a shapely point.

        Args:
            point: A shapely point
            *args: Positional arguments to be passed to :class:`~CartesianLocation`
            **kwargs: Keyword arguments to be passed to :class:`~CartesianLocation`

        Returns:
            The cartesian location created from the given geometry and other parameters
        """
        return cls(point.x, point.y, *args, **kwargs)

    def translate(self, direction: float, distance: float) -> Tuple["CartesianLocation", ndarray]:
        """Translates this location.

        Args:
            direction: The direction angle in degrees (``0`` is north, clockwise)
            distance: The distance to translate bin meters

        Returns:
            The translated polygon and the translation vector ``(x_offset, y_offset)`` in meters
            that can be used to reconstruct the original polygon
        """

        x_offset = sin(radians(direction)) * distance
        y_offset = cos(radians(direction)) * distance

        return (
            CartesianLocation.from_shapely(
                translate(Point(self.east, self.north), xoff=x_offset, yoff=y_offset),
                location_type=self.location_type,
                name=self.name,
                identifier=self.identifier,
                origin=self.origin,
            ),
            array([-x_offset, -y_offset]),
        )

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {"type": "Point", "coordinates": (self.east, self.north)}

    def __copy__(self) -> "CartesianLocation":
        return CartesianLocation(
            east=self.east,
            north=self.north,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
            origin=self.origin,
        )

    def __deepcopy__(self, memodict: Dict) -> "CartesianLocation":
        return CartesianLocation(
            east=deepcopy(self.east, memodict),
            north=deepcopy(self.north, memodict),
            location_type=deepcopy(self.location_type, memodict),
            name=deepcopy(self.name, memodict),
            identifier=deepcopy(self.identifier, memodict),
            origin=deepcopy(self.origin, memodict),
        )

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    # Inherits the docstring
    def equals(self, other: Any) -> bool:  # pylint: disable=missing-function-docstring
        return (
            isinstance(other, CartesianLocation)
            and Point.equals(self, other)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    # Inherits the docstring
    def equals_exact(  # pylint: disable=missing-function-docstring
        self, other: Any, tolerance: float
    ) -> bool:
        return (
            isinstance(other, CartesianLocation)
            and Point.equals_exact(self, other, tolerance)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    def __repr__(self) -> str:
        origin = f", origin={self.origin}" if self.origin is not None else ""
        return f"CartesianLocation(east={self.east}, north={self.north}{origin}{self._repr_extras})"

    def __str__(self) -> str:
        # this is required to override shapely.geometry.Point.__str__()
        return self.__repr__()
