"""This module implements abstractions for geospatial, polygonal shapes in WGS84 and local cartesian
coordinates using shapely.

Two polygons are ``==`` if and only if they are equal according to ``equals_exact()``.
"""

# Python standard library
from copy import deepcopy
from math import cos
from math import radians
from math import sin

# Typing
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

# Data modelling
from numpy import array
from numpy import isfinite
from numpy import ndarray
from shapely.affinity import translate
from shapely.geometry import Polygon

# Geospatial basis
from .geospatial import Geospatial
from .geospatial import LocationType

# Geospatial helpers
from .helpers import mean_coordinate
from .helpers import translate_numpy
from .helpers import WGS84_PYPROJ_GEOD

# Location representation
from .location import CartesianLocation
from .location import PolarLocation


class PolarPolygon(Geospatial):

    """A polygon based on WGS84 coordinates.

    An object with only a single point may be represented by a polygon with three times the same location.

    Args:
        locations: The points that make up this polygon; see :attr:`~.locations`
        location_type: The type of this polygon
        name: An optional name of this polygon
        identifier: The polygon's optional unique identifier, in :math:`[0, 2**63)`, i.e. 64 signed bits
    """

    def __init__(
        self,
        locations: List[PolarLocation],
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
    ) -> None:
        # Type hints
        self._locations: List[PolarLocation]

        # Attributes setup
        self.locations = locations
        self._mean_location: Optional[PolarLocation] = None

        super().__init__(location_type=location_type, name=name, identifier=identifier)

    @property
    def locations(self) -> List[PolarLocation]:
        """The points that make up this polygon.

        Getter:
            At least three points are returned.

        Setter:
            The list is closed if not already done, such that the first and last points in the list always
            match exactly. Raises an :class:`AssertionError` if less than three points are given.
        """

        return self._locations

    @locations.setter
    def locations(self, locations: List[PolarLocation]) -> None:
        assert len(locations) >= 3, "a polygon must contain at least three points"

        # close the ring as shapely would do it
        # comparison is done by exact comparison
        if (
            locations[0].latitude != locations[-1].latitude
            or locations[0].longitude != locations[-1].longitude
        ):
            locations.append(locations[0])

        self._locations = locations
        self._mean_location = None

    def to_cartesian(self, origin: PolarLocation) -> "CartesianPolygon":
        """Projects this polygon to a cartesian representation according to the given global reference.

        Args:
            origin: The reference point by which to project onto the local tangent plane

        Returns:
            The cartesian representation of this polygon with the given reference point being set
        """

        # convert to cartesian
        coordinates = self.to_numpy()
        coordinates[:, 0], coordinates[:, 1] = origin.projection(coordinates[:, 0], coordinates[:, 1])

        return CartesianPolygon.from_numpy(
            coordinates,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
            origin=origin,
        )

    def distance_to_vertices(self, location: PolarLocation, approximate: bool = False) -> float:
        """Computes the distance of the given location to the nearest vertex of this polygon.

        Args:
            location: The location to compute the distance from
            approximate: Whether to use a less precise, faster method or not
        """

        return min([location.distance(loc, approximate) for loc in self.locations])

    @property
    def area(self) -> float:
        """Returns the area of the polygon in :math:`meters^2`.

        Only simple polygons are supported, i.e. not self-intersecting ones.
        See :meth:`pyproj.Geod.polygon_area_perimeter` for the implementation.
        The returned value is always non-negative.
        """

        _, area = WGS84_PYPROJ_GEOD.polygon_area_perimeter(
            lons=[location.longitude for location in self.locations],
            lats=[location.latitude for location in self.locations],
            radians=False,
        )
        # pyproj is not typed as of now
        # the returned area is signed with the ordering of the points
        return abs(area)

    @property
    def is_valid(self) -> bool:
        """Whether this geometry is valid according to :mod:`shapely`. Quite expensive, not cached.

        Invalid ones might cross themselves or have zero area. Other tools might still refuse it, like *GEOS*.
        """

        return cast(bool, self.to_cartesian(self.mean).is_valid)

    def simplify(self, tolerance: float, preserve_topology: bool = True) -> "PolarPolygon":
        """Creates a simplified copy analogous to :meth:`shapely.geometry.Polygon.simplify`.

        The simplification is achieved by reducing its number of vertices in a way that least deforms the
        shape.

        Args:
            tolerance: This is passed to :meth:`shapely.geometry.Polygon.simplify`:
                "All points in the simplified object will be within the tolerance distance of the
                original geometry."
            preserve_topology: This is passed to :meth:`shapely.geometry.Polygon.simplify`:
                "By default a slower algorithm is used that preserves topology."

        Returns:
            A simplified version of the polygon with the same other attributes
        """

        projection_center = self.mean
        cartesian = self.to_cartesian(projection_center)

        simplified = cartesian.simplify(tolerance, preserve_topology)
        coords = array(simplified.exterior.xy).T  # this is the fastest known method

        result_cartesian = CartesianPolygon.from_numpy(
            coords, location_type=self.location_type, name=self.name, identifier=self.identifier
        )
        return result_cartesian.to_polar(projection_center)

    def translate(self, direction: float, distance: float) -> Tuple["PolarPolygon", ndarray]:
        """Translates this location and returns the new polygon and back-azimuth.

        See :func:`pyrate.plan.geometry.helpers.translate_floats` for details.
        """

        new_coordinates, back_azimuth_array = translate_numpy(self.to_numpy(), direction, distance)

        new_polygon = PolarPolygon.from_numpy(
            new_coordinates,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
        )

        return new_polygon, back_azimuth_array

    def to_numpy(self) -> ndarray:
        """Converts the coordinates defining this polygon into a :class:`numpy.ndarray`.

        Returns:
            An array with shape ``(number of locations, 2)``, where each location is represented by a pair of
            ``(longitude, latitude)``, each in degrees.

        See Also:
            :meth:`~from_numpy`
        """

        return array(
            [(location.longitude, location.latitude) for location in self.locations],
            dtype="float64",
            order="C",
        )

    @classmethod
    def from_numpy(cls, data: ndarray, *args, **kwargs) -> "PolarPolygon":
        """Create a polar polygon from a numpy representation.

        Args:
            data: An array with shape ``(number of locations, 2)``, where each location is represented by a
                  pair of ``(longitude, latitude)``, each in degrees.
            args: positional arguments to be passed to :class:`~PolarPolygon`
            kwargs: keyword arguments to be passed to :class:`~PolarPolygon`

        Returns:
            The polar polygon created from the given coordinates and other parameters

        Raises:
            AssertionError: If the shape of ``data`` is invalid or contains non-finite values

        See Also:
            :meth:`~to_numpy`
        """

        assert len(data.shape) == 2
        assert data.shape[1] == 2
        assert isfinite(data).all(), "Invalid values in CartesianPolygon.from_numpy()"

        return cls([PolarLocation(latitude=lat, longitude=lon) for (lon, lat) in data], *args, **kwargs)

    @property
    def mean(self) -> PolarLocation:
        """Computes a reasonable mean location of the polygon, if possible. The result is cached.

        Raises:
            ValueError: If no meaningful mean (of the longitudes) can be computed.
                See :func:`pyrate.plan.geometry.helpers.mean_angle`.
        """

        if self._mean_location is None:
            coordinates = self.to_numpy()
            latitude, longitude = mean_coordinate(latitudes=coordinates[:, 1], longitudes=coordinates[:, 0])
            name = f"{self.name} - mean" if self.name else "mean"
            self._mean_location = PolarLocation(latitude, longitude, name=name)

        return self._mean_location

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {
            "type": "Polygon",
            "coordinates": [
                # the inner array is only the exterior ring,
                # and we don't have an interior one
                [(location.longitude, location.latitude) for location in self.locations]
            ],
        }

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    def equals(self, other: Any) -> bool:  # pylint: disable=missing-function-docstring
        return (
            isinstance(other, PolarPolygon)
            and self.to_cartesian(self.mean).equals(other.to_cartesian(self.mean))
            and Geospatial.__eq__(self, other)
        )

    equals.__doc__ = PolarLocation.equals.__doc__

    def equals_exact(self, other: Any, tolerance: float) -> bool:
        # pylint: disable=missing-function-docstring

        return (
            isinstance(other, PolarPolygon)
            and self.to_cartesian(self.mean).equals_exact(other.to_cartesian(self.mean), tolerance)
            and Geospatial.__eq__(self, other)
        )

    equals_exact.__doc__ = PolarLocation.equals_exact.__doc__

    def equals_almost_congruent(
        self, other: Any, rel_tolerance: float = 1e-6, abs_tolerance: float = 1e-6
    ) -> bool:
        """Returns whether two objects are approximately congruent and their attributes equal exactly.

        See :meth:`~almost_congruent` for details on the specific definition of congruence and the tolerances.

        Args:
            other: The object to compare with
            rel_tolerance: The relative tolerance (relative to the larger area)
            abs_tolerance: The absolute area of tolerance in square meters

        Returns:
            Whether this and the ``other`` polygon are approximately congruent and all attributes are equal.
            Returns ``False`` if ``other`` is not a :class:`~PolarPolygon`.
        """

        return (
            isinstance(other, PolarPolygon)
            and self.almost_congruent(other, rel_tolerance=rel_tolerance, abs_tolerance=abs_tolerance)
            and Geospatial.__eq__(self, other)
        )

    def almost_congruent(
        self, other: "PolarPolygon", rel_tolerance: float = 1e-6, abs_tolerance: float = 1e-6
    ) -> bool:
        """Returns whether two polygons are approximately congruent while allowing for small differences.

        This function is not directly part of shapely and is somewhat costly to compute. It has to:
        - Project both polygons to cartesian coordinates (to continue with shapely calculations).
        - Calculate the area of the symmetric difference between this and the other polygon.
        - Calculate the area of both individual polygons.

        The arguments follow the style of :func:`math.isclose`.

        Args:
            other: The polygon to compare with
            rel_tolerance: The relative tolerance (relative to the larger area)
            abs_tolerance: The absolute area of tolerance in square meters

        Returns:
            Whether this and the other polygon are approximately congruent. The larger one of the relative
            and absolute tolerance is used.
        """

        return self.to_cartesian(self.mean).almost_congruent(
            other.to_cartesian(self.mean), rel_tolerance=rel_tolerance, abs_tolerance=abs_tolerance
        )

    def __repr__(self) -> str:
        locations = ", ".join(str(loc) for loc in self.locations)

        return f"PolarPolygon(locations=[{locations}]{self._repr_extras})"


class CartesianPolygon(Geospatial, Polygon):
    """A cartesian polygon based on local coordinates with an optional global reference.

    Note:
        For the sake of simplicity and performance, this class does not store the given
        :class:`~pyrate.plan.geometry.location.CartesianLocation` instances directly,
        but only their coordinates.
        Thus, when reading back attributes like ``origin``, ``name``, etc. of the locations they are derived
        from the polygon instance and not from the individual locations.

    Args:
        locations: The list of locations that this shape consists of; see :attr:`~.locations`
        location_type: The type of this polygon
        name: The name of this polygon
        identifier: The polygon's optional unique identifier, in :math:`[0, 2**63)`, i.e. 64 signed bits
        origin: A reference that can be used to project this cartesian representation (back)
            into a polar one
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        locations: List[CartesianLocation],
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
        origin: Optional[PolarLocation] = None,
    ) -> None:
        self.origin = origin

        if isinstance(locations, list):
            Polygon.__init__(self, [location.coords[0] for location in locations])
        else:
            # this is required for an efficient implementation of CartesianPolygon.from_numpy
            # we do not add this possibility to the type signature to make people use from_numpy().
            Polygon.__init__(self, locations)

        Geospatial.__init__(self, location_type=location_type, name=name, identifier=identifier)

    @property
    def locations(self) -> List[CartesianLocation]:
        """Get the locations of this polygon. See the class description for caveats."""

        return [
            CartesianLocation(
                x,
                y,
                origin=self.origin,
                location_type=self.location_type,
                name=self.name,
                identifier=self.identifier,
            )
            for (x, y) in self.exterior.coords
        ]

    def to_polar(self, origin: Optional[PolarLocation] = None) -> PolarPolygon:
        """Computes the polar representation of this shape.

        Args:
            origin: The global reference to be used for back-projection, must be set if and only if
                :attr:`~pyrate.plan.geometry.CartesianPolygon.origin` is ``None``

        Returns:
            The global, polar representation of this geometry
        """

        if origin is None:
            if self.origin is None:
                raise ValueError("need to give an explicit origin when the instance does not have one")
            origin = self.origin
        elif self.origin is not None:
            raise ValueError("provided an explicit origin while the instance already has one")

        # convert to cartesian
        coordinates = self.to_numpy()
        coordinates[:, 0], coordinates[:, 1] = origin.projection(
            coordinates[:, 0], coordinates[:, 1], inverse=True
        )

        return PolarPolygon.from_numpy(
            coordinates, location_type=self.location_type, name=self.name, identifier=self.identifier
        )

    def to_numpy(self) -> ndarray:
        """Converts the coordinates defining this polygon into a :class:`numpy.ndarray`.

        Returns:
            An array with shape ``(number of locations, 2)``, where each location is represented by a pair of
            ``(longitude, latitude)``, each in degrees.

        See Also:
            :meth:`~from_numpy`
        """

        return array(self.exterior.coords, dtype="float64", order="C")

    @classmethod
    def from_numpy(cls, data: ndarray, *args, **kwargs) -> "CartesianPolygon":
        """Create a cartesian polygon from a numpy representation.

        Args:
            data: An array with shape ``(number of locations, 2)``, where each location is represented by a
                  pair of ``(longitude, latitude)``, each in degrees.
            *args: Positional arguments to be passed to :class:`~CartesianPolygon`
            **kwargs: Keyword arguments to be passed to :class:`~CartesianPolygon`

        Returns:
            The polar polygon created from the given coordinates and other parameters

        Raises:
            AssertionError: If the shape of ``data`` is invalid or contains non-finite values

        See Also:
            :meth:`~to_numpy`
        """

        assert len(data.shape) == 2
        assert data.shape[1] == 2
        assert isfinite(data).all(), "Invalid values in PolarPolygon.from_numpy()"

        return cls(data, *args, **kwargs)  # type: ignore

    @classmethod
    def from_shapely(cls, polygon: Polygon, *args, **kwargs) -> "CartesianPolygon":
        """Create a cartesian polygon from a shapely polygon.

        Args:
            polygon: A shapely polygon
            *args: Positional arguments to be passed to :class:`~CartesianPolygon`
            **kwargs: Keyword arguments to be passed to :class:`~CartesianPolygon`

        Returns:
            The cartesian polygon created from the given geometry and other parameters
        """
        return cls.from_numpy(array(polygon.exterior.xy).T, *args, **kwargs)

    def translate(self, direction: float, distance: float) -> Tuple["CartesianPolygon", ndarray]:
        """Translates this polygon.

        Args:
            direction: The direction angle in degrees (``0`` is north, clockwise)
            distance: The distance to translate bin meters

        Returns:
            The translated polygon and the translation vector ``(x_offset, y_offset)`` in meters
            that can be used to reconstruct the original polygon
        """

        x_offset = sin(radians(direction)) * distance
        y_offset = cos(radians(direction)) * distance

        return (
            CartesianPolygon.from_shapely(
                translate(Polygon(self.to_numpy()), xoff=x_offset, yoff=y_offset),
                location_type=self.location_type,
                name=self.name,
                identifier=self.identifier,
                origin=self.origin,
            ),
            array([-x_offset, -y_offset]),
        )

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {
            "type": "Polygon",
            "coordinates": [
                # the inner array is only the exterior ring,
                # and we don't have an interior one
                list(self.exterior.coords),
            ],
        }

    def __copy__(self) -> "CartesianPolygon":
        return CartesianPolygon(
            locations=self.locations,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
            origin=self.origin,
        )

    def __deepcopy__(self, memodict: Dict) -> "CartesianPolygon":
        return CartesianPolygon(
            locations=deepcopy(self.locations, memodict),
            location_type=deepcopy(self.location_type, memodict),
            name=deepcopy(self.name, memodict),
            identifier=deepcopy(self.identifier, memodict),
            origin=deepcopy(self.origin, memodict),
        )

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    # Inherits the docstring
    def equals(self, other: Any) -> bool:  # pylint: disable=missing-function-docstring
        return (
            isinstance(other, CartesianPolygon)
            and Polygon.equals(self, other)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    # Inherits the docstring
    def equals_exact(  # pylint: disable=missing-function-docstring
        self, other: Any, tolerance: float
    ) -> bool:
        return (
            isinstance(other, CartesianPolygon)
            and Polygon.equals_exact(self, other, tolerance)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    def equals_almost_congruent(
        self, other: Any, rel_tolerance: float = 1e-6, abs_tolerance: float = 1e-6
    ) -> bool:
        """Returns whether two objects are approximately congruent and their attributes equal exactly.

        See :meth:`~almost_congruent` for details on the specific definition of congruence and the tolerances.

        Args:
            other: The object to compare with
            rel_tolerance: The relative tolerance (relative to the larger area)
            abs_tolerance: The absolute area of tolerance in square meters

        Returns:
            Whether this and the ``other`` polygon are approximately congruent and all attributes are equal.
            Returns ``False`` if ``other`` is not a :class:`~CartesianPolygon`.
        """

        return (
            isinstance(other, CartesianPolygon)
            and self.almost_congruent(other, rel_tolerance=rel_tolerance, abs_tolerance=abs_tolerance)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    def almost_congruent(
        self, other: "CartesianPolygon", rel_tolerance: float = 1e-6, abs_tolerance: float = 1e-6
    ) -> bool:
        """Returns whether two polygons are approximately congruent while allowing for small differences.

        This function is not directly part of shapely and is somewhat costly to compute. It has to:
        - Calculate the area of the symmetric difference between this and the ``other`` polygon.
        - Calculate the area of both individual polygons.

        The arguments follow the style of :func:`math.isclose`.

        Args:
            other: The polygon to compare with
            rel_tolerance: The relative tolerance (relative to the larger area)
            abs_tolerance: The absolute area of tolerance in square meters

        Returns:
            Whether this and the ``other`` polygon are approximately congruent. The larger one of the relative
            and absolute tolerance is used.
        """

        rel_tolerance_as_abs: float = max(self.area, other.area) * rel_tolerance
        tolerance: float = max(rel_tolerance_as_abs, abs_tolerance)
        difference: float = self.symmetric_difference(other).area
        return difference <= tolerance

    def __repr__(self) -> str:
        origin = f", origin={self.origin}" if self.origin is not None else ""
        locations = ", ".join(f"({x}, {y})" for x, y in self.exterior.coords)

        return f"CartesianPolygon(locations=[{locations}]{origin}{self._repr_extras})"

    def __str__(self) -> str:
        # this is required to override shapely.geometry.Polygon.__str__()
        return self.__repr__()
