"""
naive code for polar model used in the optimizer and cost model names
"""

# Dataclass Support
from abc import ABC
from abc import abstractmethod

# Static Typing
from typing import cast

from numpy.typing import NDArray

# Scientific Computing
import numpy as np


from scipy.interpolate import UnivariateSpline


class BoatSpeedPolar(ABC):
    """An exemplary boat speed polar.

    It provides the wind speed relation based on the boats speed polar diagram
    (wind direction and strength dependent boat speed).
    An example can be seen `Link here <https://en.wikipedia.org/wiki/Polar_diagram_(sailing)>`

    for future: improve simulation capabilities or include an interface to an simulation node
    """

    def __init__(self) -> None:
        # spline
        self.values = {"alpha": [np.pi * i / 5.0 for i in range(6)], "v": [0.9, 1.6, 1.5, 1.0, 0, 0]}

        alphas = np.array(self.values["alpha"][:-1] + [0.85 * np.pi] + [self.values["alpha"][-1]])
        self.alphas = np.concatenate((np.flipud(-alphas[1:-1]), alphas, 2 * np.pi + np.flipud(-alphas[1:-1])))

        speeds = np.array(self.values["v"][:-1] + [-0.04] + [self.values["v"][-1]])
        self.speeds = np.concatenate((np.flipud(speeds[1:-1]), speeds, np.flipud(speeds[1:-1])))
        self.speed_polar = UnivariateSpline(self.alphas, self.speeds, s=0)

    @abstractmethod
    def speed(self, angles: NDArray[np.floating]) -> NDArray[np.floating]:
        """Returns simulated speed of the boat at a given angle.

        Args:
            angles: angles to simulate speed at

        Returns:
            simulated speed
        """

    @abstractmethod
    def speed_grad(self, angles: NDArray[np.floating]) -> NDArray[np.floating]:
        """Returns gradient at a given array of angles

        Args:
            angles:  angles to simulate speed at

        Returns:
            gradient of the spline model to use in general cost gradient calculation
        """

    def _speed(
        self, angle: NDArray[np.floating], wind_speed: np.floating, wind_direction: np.floating
    ) -> NDArray[np.floating]:
        """Return the boat speed at given heading, wind speed and wind angle.

        Spline interpolation regarding angle and quadratic model for wind speed

        Args:
            angle: wind angle to evaluate spline model on
            wind_speed: wind speed to evaluate spline model on
            wind_direction: wind direction to evaluate spline model on

        Returns:
            predicted speed for the boat at given circumstance
        """
        max_speed = 15
        speed_clipped = np.clip(wind_speed, np.float64(0), np.float64(max_speed))
        return np.multiply(
            self._spline_eval(angle - wind_direction), 1 - (max_speed - speed_clipped) / max_speed
        )

    def _speed_grad(
        self, angle: NDArray[np.floating], wind_speed: np.floating, wind_direction: np.floating
    ) -> NDArray[np.floating]:
        """Return the boat speed derivative.

        Does this with respect to the angle at given heading, wind speed and wind angle
        spline interpolation regarding angle and quadratic model for wind speed

        Args:
            angle: angle to determine speed derivative from
            wind_speed: wind speed to determine speed derivative from
            wind_direction: wind direction to determine speed derivative from
        """
        max_speed = 15
        speed_clipped = np.clip(wind_speed, np.float64(0), np.float64(max_speed))
        return np.multiply(
            self._spline_der(angle - np.float32(wind_direction)), 1 - (max_speed - speed_clipped) / max_speed
        )

    def _spline_eval(self, angle: NDArray[np.floating]) -> NDArray[np.floating]:
        """Evaluate spline at a given angle.

        Args:
            angle: angle to evaluate spline at

        Returns:
            spline evaluation at given angle
        """
        ang = np.add(angle, np.pi) % (2 * np.pi) - np.pi

        return cast(NDArray[np.floating], self.speed_polar(np.abs(ang)))

    def _spline_der(self, angle: NDArray[np.floating]) -> NDArray[np.floating]:
        """Evaluates spline derivative at given angle.

        Args:
            angle: angle to evaluate spline derivative at

        Returns:
            spline derivative at given angle
        """
        ang = (angle + np.pi) % (2 * np.pi) - np.pi

        return cast(NDArray[np.floating], self.speed_polar(np.abs(ang), nu=1) * np.sign(ang))


class PolarModel(BoatSpeedPolar):
    """A very simple implementation of a polar model.

    Args:
        wind_speed: The speed of the wind in, m/s
        wind_direction: The direction from which the wind is blowing, in radians
        manoeuvre_time: approximate time each turning manoeuvre takes
    """

    def __init__(
        self, wind_speed: float = 0, wind_direction: float = 0, manoeuvre_time: float = 10.0
    ) -> None:
        super().__init__()

        self.wind_speed = np.float64(wind_speed)
        self.wind_direction = np.float64(wind_direction)
        self.manoeuvre_time = np.float64(manoeuvre_time)

    def speed(self, angles: NDArray[np.floating]) -> NDArray[np.floating]:
        """Returns simulated speed of the boat at a given angle.

        Args:
            angles: angles to simulate speed at

        Returns:
            simulated speed
        """
        return self._speed(angles, self.wind_speed, self.wind_direction)

    def speed_grad(self, angles: NDArray[np.floating]) -> NDArray[np.floating]:
        """Returns gradient at a given array of angles.

        Args:
            angles:  angles to simulate speed at

        Returns:
            gradient of the spline model to use in general cost gradient calculation
        """
        #print(f"windA {angles} {self.wind_direction}")
        _ = self._speed_grad(angles, self.wind_speed, self.wind_direction)
        #print(f"grad {_}")
        return _
