"""
utilities module for the planning backend
"""

# Scientific Computing
import numpy as np
import numpy.typing as npt

# Geometry
from shapely.geometry import Point


def angle_between(vector_a: np.ndarray, vector_b: np.ndarray) -> float:
    """Code snippet to determine right-hand angle between two numpy arrays.

    Args:
        vector_a: the first vector, of shape ``(spacial_dimension, )``
        vector_b: the second vector, of shape ``(spacial_dimension, )``

    Returns:
        the right-hand angle encompassed between ``vector_a`` and ``vector_b``, in degrees
    """

    ang1 = np.arctan2(*vector_a[::-1])
    ang2 = np.arctan2(*vector_b[::-1])
    return float(np.rad2deg((ang1 - ang2) % (2 * np.pi)))


def transform_angles_leftx_to_righty(angle: np.float64) -> np.float64:
    """Transpose angle from lefthand x axis to righthand y-axis angle (to north)

    Args:
        angle: lefthand angle measured from x axis [-pi:pi]
    Returns:
        right hand angle referenced from y axis [-pi:pi]
    """
    mapped_angle = 0.5 * np.pi - angle
    if mapped_angle > 0.5 * np.pi:
        mapped_angle = -(2 * np.pi - mapped_angle)
    return mapped_angle


def merge_numpy_array(
    array_a: npt.NDArray[np.floating], array_b: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]:
    """Merges two numpy arrays into one along the first axis.

    Args:
        array_a: numpy array a to merge
        array_b: numpy array b to merge

    Returns:
        merged numpy array
    """
    # merge two numpy arrays in alternating order along the first dimension
    if len(array_b) == 0:
        return array_a
    shape = array_a.shape[0] + array_b.shape[0]
    res = np.zeros([shape] + list(array_a.shape[1:]), dtype=np.float64)
    res[::2] = array_a
    res[1::2] = array_b

    return res


def shapely_point_to_ndarray(point: Point) -> npt.NDArray[np.floating]:
    """Projects a :mod:`shapely`/:mod:`pyrate.plan.geometry` point onto a numpy array.

    Args:
        point: point to be projected

    Returns:
        The numpy-array representation of the point in ``(x, y)`` format
    """

    return np.array([point.x, point.y], dtype=np.float64)
