"""This module implements the extended Gaussian Mixture PHD filter for linear
   multi target tracking.."""

# Standard library
from copy import deepcopy

# Typing
from typing import Callable
from typing import List
from typing import Union

# Mathematics
from numpy.linalg import inv
from numpy import ndarray

# Gaussians for state representation
from pyrate.common.math import Gaussian

# Base class
from .gmphd import GaussianMixturePHD


class ExtendedGaussianMixturePHD(GaussianMixturePHD):

    """The extended gaussian mixture PHD filter for non-linear multi-target tracking.

    The extended gaussian mixture PHD filter is a multi target tracker for non-linear state space models.
    It can be regarded as an extension of the extended Kalman filter formulas to so-called random
    finite sets (RFS). The PHD filter follows the same prediction-correction scheme for state
    estimation as the single target extended Kalman filters. As an additional part of the interface,
    the internal model for the filter's belief needs to be pruned regularly as to limit
    the computational complexity. The extraction of a state estimate is similarly more
    sophisticated in the PHD filter and requires the use of a dedicated procedure.

    Examples:
        Start by importing the necessary numpy functions.

        >>> from numpy import array
        >>> from numpy import cos
        >>> from numpy import eye
        >>> from numpy import sin
        >>> from numpy import vstack

        Setup the model. In this case, we track sine waves.
        Thereby we choose the transition model and its jacobian, as well as the linear
        measurement model, like so.

        >>> f = lambda x: sin(x)
        >>> F = lambda x: array([cos(x)])
        >>> H = lambda x: array([[1.0]])
        >>> h = lambda x: x

        Furthermore, we assume the following noise on the process and measurements.

        >>> Q = eye(1)
        >>> R = eye(1)

        Our belief of how targets are generetaded is for them to start with
        a position at zero.

        >>> mean = vstack([0.0])
        >>> covariance = array([[1.0]])
        >>> birth_belief = [Gaussian(mean, covariance)]

        We need to tell the filter how certain we are to detect targets and whether they survive.
        Also, the amount of clutter in the observed environment is quantized.

        >>> survival_rate = 0.99
        >>> detection_rate = 0.99
        >>> intensity = 0.01

        Then, we initialize the filter. This model has not input, so we ignore B.

        >>> phd = ExtendedGaussianMixturePHD(
        ...     birth_belief,
        ...     survival_rate,
        ...     detection_rate,
        ...     intensity,
        ...     F,
        ...     f,
        ...     H,
        ...     h,
        ...     Q,
        ...     R
        ... )

        We first predict with the provided model and then correct the prediction with a
        measurement, in this case of a single targets' position.

        >>> phd.predict()
        >>> phd.correct([array([5.])])

    Args:
        birth_belief: GMM of target births
        survival_rate: Probability of a target to survive a timestep
        detection_rate: Probability of a target to be detected at a timestep
        intensity: Clutter intensity
        F: Linear state transition model (n, n)
        f: Non-linear state transition model that describes the state's evolution
               from one timestep to the next
        H: Linearized measurement model, i.e. the jacobi matrix of h (m, n)
        h: Non-linear measurement model that maps a state variable into the measured space
        Q: Process noise matrix (n, n)
        R: Measurement noise matrix (m, m)
    """

    # In this context, we reproduce a common filter notation
    # pylint: disable=invalid-name
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        birth_belief: List[Gaussian],
        survival_rate: float,
        detection_rate: float,
        intensity: float,
        F: Union[ndarray, Callable[..., ndarray]],
        f: Callable[..., ndarray],
        H: Union[ndarray, Callable[..., ndarray]],
        h: Callable[..., ndarray],
        Q: ndarray,
        R: ndarray,
    ):
        # Extended filter specification
        self.f = f
        self.h = h

        # Initializes internal linear model
        super().__init__(birth_belief, survival_rate, detection_rate, intensity, F, H, Q, R)

    def predict(self, **kwargs) -> None:
        """Predict the future state."""

        # Spontaneous birth
        born = deepcopy(self.birth_belief)

        # Spawning off of existing targets
        # Not implemented at this point in time
        spawned: List[Gaussian] = []

        # Prediction for existing targets
        for component in self.gmm:
            component.x = self.f(x=component.x, **kwargs)
            F = self.F(component.x, **kwargs) if callable(self.F) else self.F
            component.P = F @ component.P @ F.T + self.Q
            component.w *= self.survival_rate

        # Concatenate with newborn and spawned target components
        self.gmm += born + spawned

    def correct(self, measurements: ndarray, **kwargs) -> None:
        """Correct the former prediction based on a sensor reading.

        Args:
            measurements: Measurements at this timestep
            **kwargs: Optional measurement models H and/or h as well as their parameters
        """

        # pylint: disable=too-many-locals

        # Check for differing measurement model
        H, h = kwargs.pop("H", self.H), kwargs.pop("h", self.h)

        # ######################################
        # Construction of update components

        mu: List[ndarray] = []  # Means mapped to measurement space
        S: List[ndarray] = []  # Residual covariance
        K: List[ndarray] = []  # Gains
        P: List[ndarray] = []  # Covariance

        for i, component in zip(range(len(self.gmm)), self.gmm):
            # Approximate about predicted state
            H_x: ndarray = H(component.x, **kwargs) if callable(H) else H

            mu.append(h(component.x, **kwargs))
            S.append(self.R + H_x @ component.P @ H_x.T)
            K.append(component.P @ H_x.T @ inv(S[i]))
            P.append(component.P - K[i] @ S[i] @ K[i].T)

        # ######################################
        # Update

        # Undetected assumption
        updated = deepcopy(self.gmm)
        for component in updated:
            component.w *= 1 - self.detection_rate

        # Measured assumption
        for z in measurements:
            # Fill batch with corrected components
            batch = [
                Gaussian(
                    self.gmm[i].x + K[i] @ (z - mu[i]),
                    P[i],
                    self.detection_rate * self.gmm[i].w * Gaussian(mu[i], S[i])(z),
                )
                for i in range(len(self.gmm))
            ]

            # Normalize weights
            sum_of_weights = sum([c.w for c in batch])
            for component in batch:
                component.w /= self.intensity + sum_of_weights

            # Append batch to updated GMM
            updated += batch

        # Set updated as new gaussian mixture model
        self.gmm = updated
