"""This module asserts correct runtime behaviour of the pyrate.sense.filter subpackage."""

# Python standard library
from datetime import timedelta

# Test environment
from unittest import TestCase

# Hypothesis testing
from hypothesis import given
from hypothesis import settings

# Package under test
from pyrate.sense.filters import ExtendedGaussianMixturePHD
from pyrate.sense.filters import GaussianMixturePHD

# Helpers
from pyrate.common.testing.strategies.dynamic_system import linear_model
from pyrate.common.testing.strategies.dynamic_system import nonlinear_model

# Flags from Pyrate
from pyrate.common.testing import IS_EXTENDED_TESTING


class TestPHD(TestCase):

    """Test for correct runtime behaviour in pyrate.sense.filter."""

    # In this context, we reproduce a common filter notation
    # pylint: disable=invalid-name

    def setUp(self) -> None:
        """Setup the linear motion model for the filter tests."""

        # Survival and detection rate of targets
        self.survival_rate = 0.99
        self.detection_rate = 0.99

        # Clutter intensity
        self.intensity = 0.01

        # PHD pruning
        self.threshold = 0.1
        self.merge_distance = 0.5
        self.max_components = 0.5

    @given(linear_model())
    @settings(max_examples=1000 if IS_EXTENDED_TESTING else 10, deadline=timedelta(seconds=1.0))
    def test_gmphd(self, model):
        """Assert the correct functionality of the gaussian mixture PHD filter."""

        # Unpack generated model
        estimate, F, _, H, Q, R, measurements, _ = model

        # Initialize filter
        gmphd = GaussianMixturePHD(
            birth_belief=[estimate],
            survival_rate=self.survival_rate,
            detection_rate=self.detection_rate,
            intensity=self.intensity,
            F=F,
            H=H,
            Q=Q,
            R=R,
        )

        self.assertEqual(len(gmphd.gmm), 0, "Mixture model should be initialy empty")

        # Predict with a callable F and check number of components
        callable_F = lambda dummy: F  # noqa: E731
        gmphd.F = callable_F
        gmphd.predict(dummy=None)
        gmphd.F = F
        self.assertEqual(
            len(gmphd.gmm), len(gmphd.birth_belief), "Mixture model is not the right size after prediction"
        )

        # Apply the PHD filter with a callable H and check number of components
        callable_H = lambda dummy: H  # noqa: E731
        gmphd.H = callable_H
        gmphd.correct(measurements=measurements, dummy=None)
        gmphd.H = H
        self.assertEqual(
            len(gmphd.gmm), len(measurements) + 1, "Mixture model is not the right size after correction"
        )

        # Removing all components
        gmphd.prune(self.threshold, self.merge_distance, 0)
        self.assertEqual(len(gmphd.extract()), 0, "Pruning did not remove all components")

        # Apply the filter multiple times
        for _, _ in enumerate(measurements):
            gmphd.predict()
            gmphd.correct(measurements=measurements)

        # Extract states
        gmphd.extract(self.threshold)

    @given(nonlinear_model())
    @settings(max_examples=1000 if IS_EXTENDED_TESTING else 10, deadline=timedelta(seconds=1.0))
    def test_extended_gmphd(self, model):
        """Assert the correct functionality of the gaussian mixture PHD filter."""

        # Unpack generated model
        estimate, f, _, Jf, h, _, Jh, Q, R, measurements = model

        # Initialize filter
        gmphd = ExtendedGaussianMixturePHD(
            birth_belief=[estimate],
            survival_rate=self.survival_rate,
            detection_rate=self.detection_rate,
            intensity=self.intensity,
            F=Jf,
            f=f,
            H=Jh,
            h=h,
            Q=Q,
            R=R,
        )

        self.assertEqual(len(gmphd.gmm), 0, "Mixture model should be initialy empty")

        # Predict and check number of components
        gmphd.predict()
        self.assertEqual(
            len(gmphd.gmm), len(gmphd.birth_belief), "Mixture model is not the right size after prediction"
        )

        # Apply the PHD filter and check number of components
        gmphd.correct(measurements=measurements)
        self.assertEqual(
            len(gmphd.gmm), len(measurements) + 1, "Mixture model is not the right size after correction"
        )

        # Removing all components
        gmphd.prune(self.threshold, self.merge_distance, 0)
        self.assertEqual(len(gmphd.extract()), 0, "Pruning did not remove all components")

        # Apply the filter multiple times
        for _, _ in enumerate(measurements):
            gmphd.predict()
            gmphd.correct(measurements=measurements)

        # Extract states
        gmphd.extract(self.threshold)
